from __future__ import annotations

from typing import TYPE_CHECKING, cast
from urllib.parse import quote

from .utils import (
    partial_pipeline_data,
    sanitize_redirect,
    setting_url,
    user_is_active,
    user_is_authenticated,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from .backends.base import BaseAuth
    from .storage import UserProtocol
    from .strategy import HttpResponseProtocol


def do_auth(backend: BaseAuth, redirect_name: str = "next") -> HttpResponseProtocol:
    # Save any defined next value into session
    data = backend.strategy.request_data(merge=False)

    # Save extra data into session.
    for field_name in cast(
        "list[str]", backend.setting("FIELDS_STORED_IN_SESSION", [])
    ):
        if field_name in data:
            backend.strategy.session_set(field_name, data[field_name])
        else:
            backend.strategy.session_set(field_name, None)

    if redirect_name in data:
        # Check and sanitize a user-defined GET/POST next field value
        redirect_uri = data[redirect_name]
        if backend.setting("SANITIZE_REDIRECTS", True):
            allowed_hosts = [
                *cast("list[str]", backend.setting("ALLOWED_REDIRECT_HOSTS", [])),
                backend.strategy.request_host(),
            ]
            redirect_uri = sanitize_redirect(allowed_hosts, redirect_uri)
        backend.strategy.session_set(
            redirect_name, redirect_uri or backend.setting("LOGIN_REDIRECT_URL")
        )
    return backend.start()


def do_complete(  # noqa: C901,PLR0912
    backend: BaseAuth,
    login: Callable,
    user: UserProtocol | None = None,
    redirect_name: str = "next",
    *args,
    **kwargs,
) -> HttpResponseProtocol:
    data = backend.strategy.request_data()

    is_authenticated = user_is_authenticated(user)
    authenticated_user: UserProtocol | HttpResponseProtocol | None = (
        user if is_authenticated else None
    )

    partial = partial_pipeline_data(backend, authenticated_user, *args, **kwargs)
    if partial:
        authenticated_user = backend.continue_pipeline(partial)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        authenticated_user = backend.complete(
            *args, user=authenticated_user, redirect_name=redirect_name, **kwargs
        )

    # pop redirect value before the session is trashed on login(), but after
    # the pipeline so that the pipeline can change the redirect if needed
    redirect_value = backend.strategy.session_get(redirect_name, "") or data.get(
        redirect_name, ""
    )

    # check if the output value is something else than a user and just
    # return it to the client
    user_model = backend.strategy.storage.user.user_model()
    if authenticated_user and not isinstance(authenticated_user, user_model):
        return cast("HttpResponseProtocol", authenticated_user)

    authenticated_user = cast("UserProtocol | None", authenticated_user)
    url: str | None

    if is_authenticated:
        if not authenticated_user:
            url = setting_url(backend, redirect_value, "LOGIN_REDIRECT_URL")
        else:
            url = setting_url(
                backend,
                redirect_value,
                "NEW_ASSOCIATION_REDIRECT_URL",
                "LOGIN_REDIRECT_URL",
            )
    elif authenticated_user:
        # check if inactive users are allowed to login
        bypass_inactivation = backend.strategy.setting(
            "ALLOW_INACTIVE_USERS_LOGIN", False
        )
        if bypass_inactivation or user_is_active(authenticated_user):
            # catch is_new/social_user in case login() resets the instance
            # These attributes are set in BaseAuth.pipeline()
            is_new = getattr(authenticated_user, "is_new", False)
            social_user = authenticated_user.social_user  # type: ignore[union-attr]
            login(backend, authenticated_user, social_user)
            # store last login backend name in session
            backend.strategy.session_set(
                "social_auth_last_login_backend", social_user.provider
            )

            if is_new:
                url = setting_url(
                    backend,
                    "NEW_USER_REDIRECT_URL",
                    redirect_value,
                    "LOGIN_REDIRECT_URL",
                )
            else:
                url = setting_url(backend, redirect_value, "LOGIN_REDIRECT_URL")
        else:
            if backend.setting("INACTIVE_USER_LOGIN", False):
                # This attribute is set in BaseAuth.pipeline()
                social_user = authenticated_user.social_user  # type: ignore[union-attr]
                login(backend, authenticated_user, social_user)
            url = setting_url(
                backend, "INACTIVE_USER_URL", "LOGIN_ERROR_URL", "LOGIN_URL"
            )
    else:
        url = setting_url(backend, "LOGIN_ERROR_URL", "LOGIN_URL")

    if not url:
        raise ValueError("By this point URL has to have been set")

    if redirect_value and redirect_value != url:
        redirect_value = quote(redirect_value)
        url += f"{'&' if '?' in url else '?'}{redirect_name}={redirect_value}"

    if backend.setting("SANITIZE_REDIRECTS", True):
        allowed_hosts = [
            *cast("list[str]", backend.setting("ALLOWED_REDIRECT_HOSTS", [])),
            backend.strategy.request_host(),
        ]
        url = sanitize_redirect(allowed_hosts, url) or backend.setting(
            "LOGIN_REDIRECT_URL"
        )
        if url is None:
            raise ValueError("Disallowed URL")
    return backend.strategy.redirect(url)


def do_disconnect(
    backend: BaseAuth,
    user: UserProtocol,
    association_id=None,
    redirect_name: str = "next",
    *args,
    **kwargs,
):
    partial = partial_pipeline_data(backend, user, *args, **kwargs)
    if partial:
        if association_id and not partial.kwargs.get("association_id"):
            partial.extend_kwargs({"association_id": association_id})
        response = backend.disconnect(*partial.args, **partial.kwargs)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        response = backend.disconnect(
            *args, user=user, association_id=association_id, **kwargs
        )

    if isinstance(response, dict):
        url: str | None = backend.strategy.absolute_uri(
            backend.strategy.request_data().get(redirect_name, "")
            or backend.setting("DISCONNECT_REDIRECT_URL")
            or backend.setting("LOGIN_REDIRECT_URL")
        )
        if backend.setting("SANITIZE_REDIRECTS", True):
            allowed_hosts = [
                *cast("list[str]", backend.setting("ALLOWED_REDIRECT_HOSTS", [])),
                backend.strategy.request_host(),
            ]
            url = (
                sanitize_redirect(allowed_hosts, url)
                or backend.setting("DISCONNECT_REDIRECT_URL")
                or backend.setting("LOGIN_REDIRECT_URL")
            )
        if not url:
            raise ValueError("Disallowed URL")
        response = backend.strategy.redirect(url)
    return response



from social_core.backends.egi_checkin import EGICheckinOpenIdConnect

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest


class EGICheckinOpenIdConnectTest(
    OpenIdConnectTest[EGICheckinOpenIdConnect], BaseAuthUrlTestMixin
):
    backend_path = "social_core.backends.egi_checkin.EGICheckinOpenIdConnect"
    issuer = "https://aai.egi.eu/auth/realms/egi"
    openid_config_body = """
    {
       "issuer": "https://aai.egi.eu/auth/realms/egi",
       "authorization_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/auth",
       "token_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/token",
       "introspection_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/token/introspect",
       "userinfo_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/userinfo",
       "end_session_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/logout",
       "frontchannel_logout_session_supported": true,
       "frontchannel_logout_supported": true,
       "jwks_uri": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/certs",
       "check_session_iframe": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/login-status-iframe.html",
       "grant_types_supported": [
           "authorization_code",
           "implicit",
           "refresh_token",
           "password",
           "client_credentials",
           "urn:ietf:params:oauth:grant-type:device_code",
           "urn:openid:params:grant-type:ciba",
           "urn:ietf:params:oauth:grant-type:token-exchange"
       ],
       "acr_values_supported": ["0", "1"],
       "response_types_supported": [
           "code",
           "none",
           "id_token",
           "token",
           "id_token token",
           "code id_token",
           "code token",
           "code id_token token"
       ],
       "subject_types_supported": ["public", "pairwise"],
       "id_token_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "id_token_encryption_alg_values_supported": [
           "RSA-OAEP",
           "RSA-OAEP-256",
           "RSA1_5"
       ],
       "id_token_encryption_enc_values_supported": [
           "A256GCM",
           "A192GCM",
           "A128GCM",
           "A128CBC-HS256",
           "A192CBC-HS384",
           "A256CBC-HS512"
       ],
       "userinfo_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512",
           "none"
       ],
       "userinfo_encryption_alg_values_supported": [
           "RSA-OAEP",
           "RSA-OAEP-256",
           "RSA1_5"
       ],
       "userinfo_encryption_enc_values_supported": [
           "A256GCM",
           "A192GCM",
           "A128GCM",
           "A128CBC-HS256",
           "A192CBC-HS384",
           "A256CBC-HS512"
       ],
       "request_object_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512",
           "none"
       ],
       "request_object_encryption_alg_values_supported": [
           "RSA-OAEP",
           "RSA-OAEP-256",
           "RSA1_5"
       ],
       "request_object_encryption_enc_values_supported": [
           "A256GCM",
           "A192GCM",
           "A128GCM",
           "A128CBC-HS256",
           "A192CBC-HS384",
           "A256CBC-HS512"
       ],
       "response_modes_supported": [
           "query",
           "fragment",
           "form_post",
           "query.jwt",
           "fragment.jwt",
           "form_post.jwt",
           "jwt"
       ],
       "registration_endpoint": "https://aai.egi.eu/auth/realms/egi/clients-registrations/openid-connect",
       "token_endpoint_auth_methods_supported": [
           "private_key_jwt",
           "client_secret_basic",
           "client_secret_post",
           "tls_client_auth",
           "client_secret_jwt"
       ],
       "token_endpoint_auth_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "introspection_endpoint_auth_methods_supported": [
           "private_key_jwt",
           "client_secret_basic",
           "client_secret_post",
           "tls_client_auth",
           "client_secret_jwt"
       ],
       "introspection_endpoint_auth_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "authorization_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "authorization_encryption_alg_values_supported": [
           "RSA-OAEP",
           "RSA-OAEP-256",
           "RSA1_5"
       ],
       "authorization_encryption_enc_values_supported": [
           "A256GCM",
           "A192GCM",
           "A128GCM",
           "A128CBC-HS256",
           "A192CBC-HS384",
           "A256CBC-HS512"
       ],
       "claims_supported": [
           "acr",
           "cert_entitlement",
           "eduperson_assurance",
           "eduperson_entitlement",
           "eduperson_scoped_affiliation",
           "eduperson_unique_id",
           "email",
           "email_verified",
           "family_name",
           "given_name",
           "name",
           "orcid",
           "preferred_username",
           "ssh_public_key",
           "sub",
           "voperson_external_affiliation",
           "voperson_id",
           "voperson_verified_email"
       ],
       "claim_types_supported": ["normal"],
       "claims_parameter_supported": true,
       "scopes_supported": [
           "openid",
           "voperson_external_affiliation",
           "email",
           "orcid",
           "aarc",
           "cert_entitlement",
           "eduperson_scoped_affiliation",
           "voperson_id",
           "ssh_public_key",
           "profile",
           "offline_access",
           "eduperson_unique_id",
           "eduperson_entitlement"
       ],
       "request_parameter_supported": true,
       "request_uri_parameter_supported": true,
       "require_request_uri_registration": true,
       "code_challenge_methods_supported": ["plain", "S256"],
       "tls_client_certificate_bound_access_tokens": true,
       "revocation_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/revoke",
       "revocation_endpoint_auth_methods_supported": [
           "private_key_jwt",
           "client_secret_basic",
           "client_secret_post",
           "tls_client_auth",
           "client_secret_jwt"
       ],
       "revocation_endpoint_auth_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "HS256",
           "HS512",
           "ES256",
           "RS256",
           "HS384",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "backchannel_logout_supported": true,
       "backchannel_logout_session_supported": true,
       "device_authorization_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/auth/device",
       "backchannel_token_delivery_modes_supported": ["poll", "ping"],
       "backchannel_authentication_request_signing_alg_values_supported": [
           "PS384",
           "ES384",
           "RS384",
           "ES256",
           "RS256",
           "ES512",
           "PS256",
           "PS512",
           "RS512"
       ],
       "require_pushed_authorization_requests": false,
       "mtls_endpoint_aliases": {
           "token_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/token",
           "revocation_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/revoke",
           "introspection_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/token/introspect",
           "device_authorization_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/auth/device",
           "registration_endpoint": "https://aai.egi.eu/auth/realms/egi/clients-registrations/openid-connect",
           "userinfo_endpoint": "https://aai.egi.eu/auth/realms/egi/protocol/openid-connect/userinfo"
       }
    }
    """

    def test_do_not_override_endpoint(self) -> None:
        self.backend.OIDC_ENDPOINT = self.issuer
        self.assertEqual(self.backend.oidc_endpoint(), self.issuer)

    def test_checkin_env_prod(self) -> None:
        self.assertEqual(
            self.backend.oidc_endpoint(), "https://aai.egi.eu/auth/realms/egi"
        )

    def test_checkin_env_demo(self) -> None:
        self.backend.CHECKIN_ENV = "demo"
        self.assertEqual(
            self.backend.oidc_endpoint(), "https://aai-demo.egi.eu/auth/realms/egi"
        )

    def test_checkin_env_dev(self) -> None:
        self.backend.CHECKIN_ENV = "dev"
        self.assertEqual(
            self.backend.oidc_endpoint(), "https://aai-dev.egi.eu/auth/realms/egi"
        )

    def test_entitlements_empty(self) -> None:
        self.assertEqual(self.backend.entitlement_allowed([]), True)

    def test_entitlements_allowed(self) -> None:
        self.backend.ALLOWED_ENTITLEMENTS = ["foo", "baz"]
        self.assertEqual(self.backend.entitlement_allowed(["foo", "bar"]), True)

    def test_entitlements_not_allowed(self) -> None:
        self.backend.ALLOWED_ENTITLEMENTS = ["baz"]
        self.assertEqual(self.backend.entitlement_allowed(["foo"]), False)



"""
Ubuntu One OpenId backend
"""

from .open_id import OpenIdAuth


class UbuntuOpenId(OpenIdAuth):
    name = "ubuntu"
    URL = "https://login.ubuntu.com"

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For Ubuntu One
        the nickname should be original.
        """
        return details["nickname"]



"""
NationBuilder OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/nationbuilder.html
"""

from typing import Any, cast

from .oauth import BaseOAuth2


class NationBuilderOAuth2(BaseOAuth2):
    """NationBuilder OAuth2 authentication backend"""

    name = "nationbuilder"
    AUTHORIZATION_URL = "https://{slug}.nationbuilder.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://{slug}.nationbuilder.com/oauth/token"
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = [("id", "id"), ("expires", "expires")]

    def get_authorization_url_format(self) -> dict[str, str]:
        return {"slug": self.slug}

    def get_access_token_url_format(self) -> dict[str, str]:
        return {"slug": self.slug}

    @property
    def slug(self) -> str:
        return cast("str", self.setting("SLUG"))

    def get_user_details(self, response):
        """Return user details from Github account"""
        email = response.get("email") or ""
        username = email.split("@")[0] if email else ""
        return {
            "username": username,
            "email": email,
            "fullname": response.get("full_name") or "",
            "first_name": response.get("first_name") or "",
            "last_name": response.get("last_name") or "",
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        url = f"https://{self.slug}.nationbuilder.com/api/v1/people/me"
        return self.get_json(url, params={"access_token": access_token})["person"]



from __future__ import annotations

import datetime
from typing import Any

from .oauth import BaseOAuth2


class AsanaOAuth2(BaseOAuth2):
    name = "asana"
    AUTHORIZATION_URL = "https://app.asana.com/-/oauth_authorize"
    ACCESS_TOKEN_URL = "https://app.asana.com/-/oauth_token"
    REFRESH_TOKEN_URL = "https://app.asana.com/-/oauth_token"
    REDIRECT_STATE = False
    USER_DATA_URL = "https://app.asana.com/api/1.0/users/me"
    EXTRA_DATA = [
        ("expires_in", "expires_in"),
        ("refresh_token", "refresh_token"),
        ("name", "name"),
    ]

    def get_user_details(self, response):
        data = response["data"]
        fullname, first_name, last_name = self.get_user_names(data["name"])
        return {
            "email": data["email"],
            "username": data["email"],
            "fullname": fullname,
            "last_name": last_name,
            "first_name": first_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            self.USER_DATA_URL, headers={"Authorization": f"Bearer {access_token}"}
        )

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        data = super().extra_data(user, uid, response, details, pipeline_kwargs)
        if self.setting("ESTIMATE_EXPIRES_ON"):
            expires_on = datetime.datetime.now(
                datetime.timezone.utc
            ) + datetime.timedelta(seconds=data["expires"])
            data["expires_on"] = expires_on.isoformat()
        return data



"""
Odnoklassniki OAuth2 and Iframe Application backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/odnoklassnikiru.html
"""

from __future__ import annotations

from hashlib import md5
from typing import Any, cast
from urllib.parse import unquote

from social_core.exceptions import AuthFailed

from .base import BaseAuth
from .oauth import BaseOAuth2


def odnoklassniki_sig(payload: str) -> str:
    """
    Calculates a payload signature using md5.

    See
    https://apiok.ru/en/ext/invite_suggest#calculating-request-signature-stsignature
    """
    return md5(payload.encode("utf-8")).hexdigest()  # noqa: S324


class OdnoklassnikiOAuth2(BaseOAuth2):
    """Odnoklassniki authentication backend"""

    name = "odnoklassniki-oauth2"
    ID_KEY = "uid"
    SCOPE_SEPARATOR = ";"
    AUTHORIZATION_URL = "https://connect.ok.ru/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.ok.ru/oauth/token.do"
    EXTRA_DATA = [("refresh_token", "refresh_token"), ("expires_in", "expires_in")]

    def get_user_details(self, response):
        """Return user details from Odnoklassniki request"""
        fullname, first_name, last_name = self.get_user_names(
            fullname=unquote(response["name"]),
            first_name=unquote(response["first_name"]),
            last_name=unquote(response["last_name"]),
        )
        return {
            "username": response["uid"],
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Return user data from Odnoklassniki REST API"""
        data = {"access_token": access_token, "method": "users.getCurrentUser"}
        _key, secret = self.get_key_and_secret()
        public_key = self.setting("PUBLIC_NAME")
        return odnoklassniki_api(
            self, data, "https://api.ok.ru/", public_key, secret, "oauth"
        )


class OdnoklassnikiApp(BaseAuth):
    """Odnoklassniki iframe app authentication backend"""

    name = "odnoklassniki-app"
    ID_KEY = "uid"

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        return {
            key: value
            for key, value in response.items()
            if key in response["extra_data_list"]
        }

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            fullname=unquote(response["name"]),
            first_name=unquote(response["first_name"]),
            last_name=unquote(response["last_name"]),
        )
        return {
            "username": response["uid"],
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def auth_complete(self, *args, **kwargs):
        self.verify_auth_sig()
        response = self.get_response()
        fields = (
            "uid",
            "first_name",
            "last_name",
            "name",
            *cast("tuple[str, ...]", self.setting("EXTRA_USER_DATA_LIST", ())),
        )
        data = {
            "method": "users.getInfo",
            "uids": str(response["logged_user_id"]),
            "fields": ",".join(fields),
        }
        _client_key, client_secret = self.get_key_and_secret()
        public_key = self.setting("PUBLIC_NAME")
        details = odnoklassniki_api(
            self,
            data,
            response["api_server"],
            public_key,
            client_secret,
            "iframe_nosession",
        )
        if len(details) == 1 and "uid" in details[0]:
            details = details[0]
            auth_data_fields = cast(
                "tuple[str, ...]",
                self.setting(
                    "EXTRA_AUTH_DATA_LIST",
                    (
                        "api_server",
                        "apiconnection",
                        "session_key",
                        "authorized",
                        "session_secret_key",
                    ),
                ),
            )

            for field in auth_data_fields:
                details[field] = response[field]
            details["extra_data_list"] = fields + auth_data_fields
            kwargs.update({"backend": self, "response": details})
        else:
            raise AuthFailed(self, "Cannot get user details: API error")
        return self.strategy.authenticate(*args, **kwargs)

    def get_auth_sig(self):
        return odnoklassniki_sig(
            f"{self.data['logged_user_id']}{self.data['session_key']}{self.setting('SECRET')}"
        )

    def get_response(self):
        fields = (
            "logged_user_id",
            "api_server",
            "application_key",
            "session_key",
            "session_secret_key",
            "authorized",
            "apiconnection",
        )
        return {name: self.data[name] for name in fields if name in self.data}

    def verify_auth_sig(self) -> None:
        correct_key = self.get_auth_sig()
        key = self.data["auth_sig"].lower()
        if correct_key != key:
            raise AuthFailed(self, "Wrong authorization key")


def odnoklassniki_oauth_sig(data, client_secret):
    """
    Calculates signature of request data access_token value must be included
    Algorithm is described at
        https://apiok.ru/wiki/pages/viewpage.action?pageId=12878032,
    search for "little bit different way"
    """
    suffix = odnoklassniki_sig(f"{data['access_token']}{client_secret}")
    check_list = sorted(
        f"{key:s}={value:s}" for key, value in data.items() if key != "access_token"
    )
    return odnoklassniki_sig("".join(check_list) + suffix)


def odnoklassniki_iframe_sig(data, client_secret_or_session_secret):
    """
    Calculates signature as described at:
        https://apiok.ru/wiki/display/ok/Authentication+and+Authorization
    If API method requires session context, request is signed with session
    secret key. Otherwise it is signed with application secret key
    """
    param_list = sorted(f"{key:s}={value:s}" for key, value in data.items())
    return odnoklassniki_sig("".join(param_list) + client_secret_or_session_secret)


def odnoklassniki_api(
    backend, data, api_url, public_key, client_secret, request_type="oauth"
):
    """Calls Odnoklassniki REST API method
    https://apiok.ru/wiki/display/ok/Odnoklassniki+Rest+API"""
    data.update({"application_key": public_key, "format": "JSON"})
    if request_type == "oauth":
        data["sig"] = odnoklassniki_oauth_sig(data, client_secret)
    elif request_type == "iframe_session":
        data["sig"] = odnoklassniki_iframe_sig(data, data["session_secret_key"])
    elif request_type == "iframe_nosession":
        data["sig"] = odnoklassniki_iframe_sig(data, client_secret)
    else:
        msg = "Unknown request type {0}. How should it be signed?"
        raise AuthFailed(backend, msg.format(request_type))
    return backend.get_json(f"{api_url}fb.do", params=data)



"""
Openshift OAuth2 backend
"""

from typing import Any, cast
from urllib.parse import urljoin

from social_core.utils import append_slash

from .oauth import BaseOAuth2


class OpenshiftOAuth2(BaseOAuth2):
    name = "openshift"

    def access_token_url(self):
        return urljoin(append_slash(cast("str", self.setting("URL"))), "oauth/token")

    def authorization_url(self):
        return urljoin(
            append_slash(cast("str", self.setting("URL"))), "oauth/authorize"
        )

    def get_user_id(self, details, response):
        return response["metadata"]["uid"]

    def get_user_details(self, response):
        """Return user details from openshift account"""
        username = response["metadata"]["name"]
        email = response["metadata"]["name"]
        return {"username": username, "email": email}

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        headers = {"Authorization": f"Bearer {access_token}"}

        return self.request(
            urljoin(append_slash(cast("str", self.setting("URL"))), "oapi/v1/users/~"),
            headers=headers,
        ).json()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DailymotionOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.dailymotion.DailymotionOAuth2"
    user_data_url = "https://api.dailymotion.com/auth/"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps({"id": "foobar", "screenname": "foobar"})

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DigitalOceanOAuthTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.digitalocean.DigitalOceanOAuth"
    user_data_url = "https://api.digitalocean.com/v2/account"
    expected_username = "sammy@digitalocean.com"
    access_token_body = json.dumps(
        {
            "access_token": "547cac21118ae7",
            "token_type": "bearer",
            "expires_in": 2592000,
            "refresh_token": "00a3aae641658d",
            "scope": "read write",
            "info": {"name": "Sammy Shark", "email": "sammy@digitalocean.com"},
        }
    )
    user_data_body = json.dumps(
        {
            "account": {
                "droplet_limit": 25,
                "email": "sammy@digitalocean.com",
                "uuid": "b6fr89dbf6d9156cace5f3c78dc9851d957381ef",
                "email_verified": True,
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Ping Auth OpenID Connect backend
"""

from typing import cast

import jwt
from jwt import (
    ExpiredSignatureError,
    InvalidAudienceError,
    InvalidTokenError,
    PyJWTError,
)
from jwt.utils import base64url_decode

from social_core.backends.open_id_connect import OpenIdConnectAuth
from social_core.exceptions import AuthTokenError


class PingOpenIdConnect(OpenIdConnectAuth):
    name = "ping"
    # OIDC_ENDPOINT has the form 'https://auth.pingone.com/<APP ID>/as'
    OIDC_ENDPOINT = ""
    REDIRECT_STATE = False
    RESPONSE_TYPE = "code"
    USERNAME_KEY = "preferred_username"

    def find_valid_key(self, id_token):
        for key in self.get_jwks_keys():
            if "alg" not in key:
                key["alg"] = "RS256"
            rsakey = jwt.PyJWK(key)
            message, encoded_sig = id_token.rsplit(".", 1)
            decoded_sig = base64url_decode(encoded_sig.encode("utf-8"))
            if rsakey.Algorithm.verify(
                message.encode("utf-8"), rsakey.key, decoded_sig
            ):
                return key
        return None

    def validate_and_return_id_token(self, id_token, access_token):
        """
        Validates the id_token according to the steps at
        http://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation.
        """
        client_id, _client_secret = self.get_key_and_secret()

        key = self.find_valid_key(id_token)

        if not key:
            raise AuthTokenError(self, "Signature verification failed")

        if "alg" not in key:
            key["alg"] = "RS256"
        rsakey = jwt.PyJWK(key)

        try:
            claims = jwt.decode(
                id_token,
                rsakey.key,
                algorithms=self.JWT_ALGORITHMS,
                audience=client_id,
                issuer=self.id_token_issuer(),
                options=self.JWT_DECODE_OPTIONS,
                leeway=cast("int", self.setting("JWT_LEEWAY", self.JWT_LEEWAY)),
            )
        except ExpiredSignatureError as error:
            raise AuthTokenError(self, "Signature has expired") from error
        except InvalidAudienceError as error:
            # compatibility with jose error message
            raise AuthTokenError(self, "Invalid audience") from error
        except InvalidTokenError as error:
            raise AuthTokenError(self, str(error)) from error
        except PyJWTError as error:
            raise AuthTokenError(self, "Invalid signature") from error

        self.validate_claims(claims)

        return claims

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", default=self.USERNAME_KEY)
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("given_name"), last_name=response.get("family_name")
        )
        return {
            "username": response.get(username_key),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



"""
Flickr OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/flickr.html
"""

from typing import Any

from .oauth import BaseOAuth1


class FlickrOAuth(BaseOAuth1):
    """Flickr OAuth authentication backend"""

    name = "flickr"
    AUTHORIZATION_URL = "https://www.flickr.com/services/oauth/authorize"
    REQUEST_TOKEN_URL = "https://www.flickr.com/services/oauth/request_token"
    ACCESS_TOKEN_URL = "https://www.flickr.com/services/oauth/access_token"
    EXTRA_DATA = [("id", "id"), ("username", "username"), ("expires", "expires")]

    def get_user_details(self, response):
        """Return user details from Flickr account"""
        fullname, first_name, last_name = self.get_user_names(response.get("fullname"))
        return {
            "username": response.get("username") or response.get("id"),
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: dict, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return {
            "id": access_token["user_nsid"],
            "username": access_token["username"],
            "fullname": access_token.get("fullname", ""),
        }

    def auth_extra_arguments(self):
        params = super().auth_extra_arguments() or {}
        if "perms" not in params:
            params["perms"] = "read"
        return params



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class FlatOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.flat.FlatOAuth2"
    user_data_url = "https://api.flat.io/v2/me"
    expected_username = "vincent"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "id": "541a137946fd04d57cb2e3c0",
            "username": "vincent",
            "name": "Vincent Foo",
            "printableName": "Vincent Foo",
            "bio": "Foo bio",
            "instruments": [],
            "picture": "https://flat-prod-public.s3.amazonaws.com/00000000/a0d2cb86-ab1e-4fdb-9286-dd94aa6d386c.jpeg",
            "registrationDate": "2014-09-17T23:04:25.042Z",
            "htmlUrl": "https://flat.io/vincent",
            "starredScoresCount": 85,
            "likedScoresCount": 85,
            "followersCount": 183,
            "followingCount": 52,
            "ownedPublicScoresCount": 15,
            "isPowerUser": True,
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Seznam OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/seznam.html
"""

from typing import Any

from .oauth import BaseOAuth2


class SeznamOAuth2(BaseOAuth2):
    """Seznam OAuth authentication backend"""

    name = "seznam-oauth2"
    API_URL = "https://login.szn.cz/api/v1/user"
    AUTHORIZATION_URL = "https://login.szn.cz/api/v1/oauth/auth"
    ACCESS_TOKEN_URL = "https://login.szn.cz/api/v1/oauth/token"
    ID_KEY = "oauth_user_id"
    STATE_PARAMETER = True
    DEFAULT_SCOPE = ["identity"]

    def api_url(self):
        return self.setting("API_URL") or self.API_URL

    def get_user_details(self, response):
        """Return user details from Seznam account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("name"),
            first_name=response.get("firstname"),
            last_name=response.get("lastname"),
        )
        return {
            "username": response.get("username"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            self.api_url(), headers={"Authorization": f"bearer {access_token}"}
        )



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""

from typing import Any, cast
from urllib.parse import urljoin, urlparse, urlunparse

from social_core.utils import append_slash

from .oauth import BaseOAuth2


class OktaMixin(BaseOAuth2):
    def api_url(self) -> str:
        return append_slash(cast("str", self.setting("API_URL")))

    def authorization_url(self):
        return self._url("v1/authorize")

    def access_token_url(self):
        return self._url("v1/token")

    def _url(self, path):
        return urljoin(self.api_url(), path)

    def oidc_config_url(self):
        # https://developer.okta.com/docs/reference/api/oidc/#well-known-openid-configuration
        url = urlparse(self.api_url())

        # If the URL path does not contain an authorizedServerId, we need
        # to truncate the path in order to generate a proper openid-configuration
        # URL.
        if url.path == "/oauth2/":
            url = url._replace(path="")

        return urljoin(
            urlunparse(url),
            f"./.well-known/openid-configuration?client_id={self.setting('KEY')}",
        )

    def oidc_config(self):
        return self.get_json(self.oidc_config_url())


class OktaOAuth2(OktaMixin, BaseOAuth2):
    """Okta OAuth authentication backend"""

    name = "okta-oauth2"
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = " "
    ID_KEY = "preferred_username"

    DEFAULT_SCOPE = ["openid", "profile", "email"]
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("expires_in", "expires_in"),
        ("token_type", "token_type", True),
    ]

    def get_user_details(self, response):
        """Return user details from Okta account"""
        return {
            "username": response.get("preferred_username"),
            "email": response.get("email") or "",
            "first_name": response.get("given_name"),
            "last_name": response.get("family_name"),
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from Okta"""
        return self.get_json(
            self._url("v1/userinfo"),
            headers={
                "Authorization": f"Bearer {access_token}",
            },
        )



from typing import cast

from social_core.exceptions import AuthMissingParameter

from .base import BaseAuth


class LegacyAuth(BaseAuth):
    def auth_url(self) -> str:
        return cast("str", self.setting("FORM_URL"))

    def auth_html(self) -> str:
        return self.strategy.render_html(tpl=cast("str", self.setting("FORM_HTML")))

    def uses_redirect(self) -> bool:
        return bool(cast("str", self.setting("FORM_URL"))) and not cast(
            "str", self.setting("FORM_HTML")
        )

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        id_key = self.id_key()
        if id_key not in self.data:
            raise AuthMissingParameter(self, id_key)
        kwargs.update({"response": self.data, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_user_details(self, response):
        """Return user details"""
        email = response.get("email", "")
        username = response.get("username", "")
        fullname, first_name, last_name = self.get_user_names(
            response.get("fullname", ""),
            response.get("first_name", ""),
            response.get("last_name", ""),
        )
        if email and not username:
            username = email.split("@", 1)[0]
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



"""
Google App Engine support using User API
"""

# pylint: disable-next=import-error,no-name-in-module
from google.appengine.api import users  # ty: ignore[unresolved-import]

from social_core.exceptions import AuthException

from .base import BaseAuth


class GoogleAppEngineAuth(BaseAuth):
    """GoogleAppengine authentication backend"""

    name = "google-appengine"

    def get_user_id(self, details, response):
        """Return current user id."""
        user = users.get_current_user()
        if user:
            return user.user_id()
        return None

    def get_user_details(self, response):
        """Return user basic information (id and email only)."""
        user = users.get_current_user()
        return {
            "username": user.user_id(),
            "email": user.email(),
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }

    def auth_url(self):
        """Build and return complete URL."""
        return users.create_login_url(self.redirect_uri)

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance."""
        if not users.get_current_user():
            raise AuthException(self, "Authentication error")
        kwargs.update({"response": "", "backend": self})
        return self.strategy.authenticate(*args, **kwargs)



"""
EVE Online Single Sign-On (SSO) OAuth2 backend
Documentation at https://eveonline-third-party-documentation.readthedocs.io/en/latest/sso/index.html
"""

from typing import Any, cast

from .oauth import BaseOAuth2


class EVEOnlineOAuth2(BaseOAuth2):
    """EVE Online OAuth authentication backend"""

    name = "eveonline"
    BASE_URL = "https://login.eveonline.com/oauth"
    AUTHORIZATION_URL = f"{BASE_URL}/authorize"
    ACCESS_TOKEN_URL = f"{BASE_URL}/token"
    ID_KEY = "CharacterID"
    EXTRA_DATA = [
        ("CharacterID", "id"),
        ("expires_in", "expires_in"),
        ("CharacterOwnerHash", "owner_hash", True),
        ("refresh_token", "refresh_token", True),
    ]

    def get_user_details(self, response):
        """Return user details from EVE Online account"""
        user_data = cast("dict", self.user_data(response["access_token"]))
        fullname, first_name, last_name = self.get_user_names(
            user_data["CharacterName"]
        )
        return {
            "email": "",
            "username": fullname,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Get Character data from EVE server"""
        return self.get_json(
            "https://login.eveonline.com/oauth/verify",
            headers={"Authorization": f"Bearer {access_token}"},
        )



from typing import Any

from .oauth import BaseOAuth2


class MonzoOAuth2(BaseOAuth2):
    """
    Monzo OAuth2 authentication backend.
    """

    name = "monzo"

    AUTHORIZATION_URL = "https://auth.getmondo.co.uk/"
    ACCESS_TOKEN_URL = "https://api.monzo.com/oauth2/token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response["accounts"][0]["description"],
        )

        return {
            "username": str(response.get("user_id")),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            "https://api.monzo.com/accounts",
            headers={"Authorization": f"Bearer {access_token}"},
        )



"""
Bitbucket Data Center OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/bitbucket_datacenter_oauth2.html
    https://confluence.atlassian.com/bitbucketserver/bitbucket-oauth-2-0-provider-api-1108483661.html
"""

from typing import Any

from .oauth import BaseOAuth2PKCE


class BitbucketDataCenterOAuth2(BaseOAuth2PKCE):
    """
    Implements client for Bitbucket Data Center OAuth 2.0 provider API.
    ref: https://confluence.atlassian.com/bitbucketserver/bitbucket-oauth-2-0-provider-api-1108483661.html
    """

    name = "bitbucket-datacenter-oauth2"
    ID_KEY = "id"
    SCOPE_SEPARATOR = " "
    REFRESH_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    # ref: https://confluence.atlassian.com/bitbucketserver/bitbucket-oauth-2-0-provider-api-1108483661.html#BitbucketOAuth2.0providerAPI-scopes
    DEFAULT_SCOPE = ["PUBLIC_REPOS"]
    USE_BASIC_AUTH = False
    EXTRA_DATA = [
        ("token_type", "token_type"),
        ("access_token", "access_token"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
        ("scope", "scope"),
        # extra user profile fields
        ("first_name", "first_name"),
        ("last_name", "last_name"),
        ("email", "email"),
        ("name", "name"),
        ("username", "username"),
        ("display_name", "display_name"),
        ("type", "type"),
        ("active", "active"),
        ("url", "url"),
        ("avatar_url", "avatar_url"),
    ]
    PKCE_DEFAULT_CODE_CHALLENGE_METHOD = "s256"  # can be "plain" or "s256"
    PKCE_DEFAULT_CODE_VERIFIER_LENGTH = 48  # must be b/w 43-127 chars
    DEFAULT_USE_PKCE = True
    DEFAULT_USER_AVATAR_SIZE = 48

    @property
    def server_base_oauth2_api_url(self) -> str:
        base_url = self.setting("URL")
        return f"{base_url}/rest/oauth2/latest"

    @property
    def server_base_rest_api_url(self) -> str:
        base_url = self.setting("URL")
        return f"{base_url}/rest/api/latest"

    def authorization_url(self) -> str:
        return f"{self.server_base_oauth2_api_url}/authorize"

    def access_token_url(self) -> str:
        return f"{self.server_base_oauth2_api_url}/token"

    def get_user_details(self, response) -> dict:
        """Return user details for the Bitbucket Data Center account"""
        # `response` here is the return value of `user_data` method
        user_data = response
        _, first_name, last_name = self.get_user_names(user_data["displayName"])
        uid = self.get_user_id(details=None, response=response)
        return {
            "uid": uid,
            "first_name": first_name,
            "last_name": last_name,
            "email": user_data["emailAddress"],
            "name": user_data["name"],
            "username": user_data["slug"],
            "display_name": user_data["displayName"],
            "type": user_data["type"],
            "active": user_data["active"],
            "url": user_data["links"]["self"][0]["href"],
            "avatar_url": user_data["avatarUrl"],
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Fetch user data from Bitbucket Data Center REST API"""
        # At this point, we don't know the current user's username
        # and Bitbucket doesn't provide any API to do so.
        # However, the current user's username is sent in every response header.
        # ref: https://community.developer.atlassian.com/t/obtain-authorised-users-username-from-api/24422/2
        headers = {"Authorization": f"Bearer {access_token}"}
        response = self.request(
            url=f"{self.server_base_rest_api_url}/application-properties",
            method="GET",
            headers=headers,
        )
        # ref: https://developer.atlassian.com/server/bitbucket/rest/v815/api-group-system-maintenance/#api-api-latest-users-userslug-get
        username = response.headers["x-ausername"]
        return self.get_json(
            url=f"{self.server_base_rest_api_url}/users/{username}",
            headers=headers,
            params={
                "avatarSize": self.setting(
                    "USER_AVATAR_SIZE", default=self.DEFAULT_USER_AVATAR_SIZE
                )  # to force `avatarUrl` in response
            },
        )



"""
Instagram OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/instagram.html
"""

from typing import Any

from .oauth import BaseOAuth2


class InstagramOAuth2(BaseOAuth2):
    name = "instagram"
    AUTHORIZATION_URL = "https://api.instagram.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.instagram.com/oauth/access_token"

    def get_user_id(self, details, response):
        user = response.get("user") or {}
        return user.get("id")

    def get_user_details(self, response):
        """Return user details from Instagram account"""
        user = response.get("user") or {}
        username = user["username"]
        email = user.get("email", "")
        fullname, first_name, last_name = self.get_user_names(user.get("full_name", ""))
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        # For more fields see:
        # https://developers.facebook.com/docs/instagram-basic-display-api/reference/user#fields
        # In fact there are not very many of them.
        fields = "id,username"
        params = {"access_token": access_token, "fields": fields}
        response = self.get_json("https://graph.instagram.com/me", params=params)
        return {"user": response}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from openid.consumer.consumer import CANCEL, FAILURE, SUCCESS, Consumer
from openid.consumer.discover import DiscoveryFailure
from openid.extensions import ax, pape, sreg
from openid.fetchers import HTTPFetchingError

from social_core.exceptions import (
    AuthCanceled,
    AuthException,
    AuthFailed,
    AuthMissingParameter,
    AuthUnknownError,
    AuthUnreachableProvider,
)
from social_core.utils import url_add_parameters

from .base import BaseAuth

if TYPE_CHECKING:
    from social_core.store import OpenIdStore

# OpenID configuration
OLD_AX_ATTRS = [
    ("http://schema.openid.net/contact/email", "old_email"),
    ("http://schema.openid.net/namePerson", "old_fullname"),
    ("http://schema.openid.net/namePerson/friendly", "old_nickname"),
]
AX_SCHEMA_ATTRS = [
    # Request both the full name and first/last components since some
    # providers offer one but not the other.
    ("http://axschema.org/contact/email", "email"),
    ("http://axschema.org/namePerson", "fullname"),
    ("http://axschema.org/namePerson/first", "first_name"),
    ("http://axschema.org/namePerson/last", "last_name"),
    ("http://axschema.org/namePerson/friendly", "nickname"),
]
SREG_ATTR = [("email", "email"), ("fullname", "fullname"), ("nickname", "nickname")]
OPENID_ID_FIELD = "openid_identifier"
SESSION_NAME = "openid"


class OpenIdAuth(BaseAuth):
    """Generic OpenID authentication backend"""

    name = "openid"
    URL: str | None = None
    USERNAME_KEY = "username"

    _consumer = None

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response.identity_url

    def get_ax_attributes(self) -> list[tuple[str, str]]:
        attrs = cast("list[tuple[str, str]]", self.setting("AX_SCHEMA_ATTRS", []))
        if attrs and self.setting("IGNORE_DEFAULT_AX_ATTRS", True):
            return attrs
        return attrs + AX_SCHEMA_ATTRS + OLD_AX_ATTRS

    def get_sreg_attributes(self):
        return self.setting("SREG_ATTR") or SREG_ATTR

    def values_from_response(self, response, sreg_names=None, ax_names=None):
        """Return values from SimpleRegistration response or
        AttributeExchange response if present.

        @sreg_names and @ax_names must be a list of name and aliases
        for such name. The alias will be used as mapping key.
        """
        values = {}

        # Use Simple Registration attributes if provided
        if sreg_names:
            # pyright does not detect the classmethod correctly
            resp = sreg.SRegResponse.fromSuccessResponse(response)  # type: ignore[reportCallIssue]
            if resp:
                values.update(
                    (alias, resp.get(name) or "") for name, alias in sreg_names
                )

        # Use Attribute Exchange attributes if provided
        if ax_names:
            # pyright does not detect the classmethod correctly
            resp = ax.FetchResponse.fromSuccessResponse(response)  # type: ignore[reportCallIssue]
            if resp:
                for src, alias in ax_names:
                    name = alias.replace("old_", "")
                    values[name] = resp.getSingle(src, "") or values.get(name)

        return values

    def get_user_details(self, response):
        """Return user details from an OpenID request"""
        values = {
            "username": "",
            "email": "",
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }
        # update values using SimpleRegistration or AttributeExchange
        # values
        values.update(
            self.values_from_response(
                response, self.get_sreg_attributes(), self.get_ax_attributes()
            )
        )

        fullname = values.get("fullname") or ""
        first_name = values.get("first_name") or ""
        last_name = values.get("last_name") or ""
        email = values.get("email") or ""

        if not fullname and first_name and last_name:
            fullname = f"{first_name} {last_name}"
        elif fullname:
            try:
                first_name, last_name = fullname.rsplit(" ", 1)
            except ValueError:
                last_name = fullname

        username_key = self.setting("USERNAME_KEY") or self.USERNAME_KEY
        values.update(
            {
                "fullname": fullname,
                "first_name": first_name,
                "last_name": last_name,
                "username": values.get(username_key)
                or (first_name.title() + last_name.title()),
                "email": email,
            }
        )
        return values

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Return defined extra data names to store in extra_data field.
        Settings will be inspected to get more values names that should be
        stored on extra_data field. Setting name is created from current
        backend name (all uppercase) plus _SREG_EXTRA_DATA and
        _AX_EXTRA_DATA because values can be returned by SimpleRegistration
        or AttributeExchange schemas.

        Both list must be a value name and an alias mapping similar to
        SREG_ATTR, OLD_AX_ATTRS or AX_SCHEMA_ATTRS
        """
        sreg_names = self.setting("SREG_EXTRA_DATA")
        ax_names = self.setting("AX_EXTRA_DATA")
        values = self.values_from_response(response, sreg_names, ax_names)
        from_details = super().extra_data(user, uid, {}, details, pipeline_kwargs)
        values.update(from_details)
        return values

    def get_return_to(self) -> str:
        params: dict[str, str] = {}
        if session_id := self.strategy.get_session_id():
            params[self.strategy.SESSION_SAVE_KEY] = session_id

        return url_add_parameters(self.strategy.absolute_uri(self.redirect_uri), params)

    def auth_url(self):
        """Return auth URL returned by service"""
        openid_request = self.setup_request(self.auth_extra_arguments())
        # Construct completion URL, including page we should redirect to
        return openid_request.redirectURL(self.trust_root(), self.get_return_to())

    def auth_html(self):
        """Return auth HTML returned by service"""
        openid_request = self.setup_request(self.auth_extra_arguments())
        form_tag = {"id": "openid_message"}
        return openid_request.htmlMarkup(
            self.trust_root(), self.get_return_to(), form_tag_attrs=form_tag
        )

    def trust_root(self):
        """Return trust-root option"""
        return self.setting("OPENID_TRUST_ROOT") or self.strategy.absolute_uri("/")

    def continue_pipeline(self, partial):
        """Continue previous halted pipeline"""
        response = self.consumer().complete(
            dict(self.data.items()), self.get_return_to()
        )
        return self.strategy.authenticate(
            self,
            *partial.args,
            response=response,
            pipeline_index=partial.next_step,
            **partial.kwargs,
        )

    def auth_complete(self, *args, **kwargs):
        """Complete auth process"""
        try:
            response = self.consumer().complete(
                dict(self.data.items()), self.get_return_to()
            )
        except HTTPFetchingError as error:
            raise AuthUnreachableProvider(self) from error

        self.process_error(response)
        if session_id := self.data.get(self.strategy.SESSION_SAVE_KEY):
            self.strategy.restore_session(session_id, kwargs)
        return self.strategy.authenticate(self, *args, response=response, **kwargs)

    def process_error(self, data) -> None:
        if not data:
            raise AuthException(self, "OpenID relying party endpoint")
        if data.status == FAILURE:
            raise AuthFailed(self, data.message)
        if data.status == CANCEL:
            raise AuthCanceled(self)
        if data.status != SUCCESS:
            raise AuthUnknownError(self, data.status)

    def setup_request(self, params=None):
        """Setup request"""
        request = self.openid_request(params)
        # Request some user details. Use attribute exchange if provider
        # advertises support.
        if request.endpoint.supportsType(ax.AXMessage.ns_uri):
            fetch_request = ax.FetchRequest()
            # Mark all attributes as required, Google ignores optional ones
            for attr, alias in self.get_ax_attributes():
                fetch_request.add(ax.AttrInfo(attr, alias=alias, required=True))
        else:
            fetch_request = sreg.SRegRequest(
                optional=list(dict(self.get_sreg_attributes()).keys())
            )
        request.addExtension(fetch_request)

        # Add PAPE Extension for if configured
        preferred_policies = self.setting("OPENID_PAPE_PREFERRED_AUTH_POLICIES")
        preferred_level_types = self.setting("OPENID_PAPE_PREFERRED_AUTH_LEVEL_TYPES")
        max_age = self.setting("OPENID_PAPE_MAX_AUTH_AGE")
        if max_age is not None:
            try:
                max_age = int(max_age)
            except (ValueError, TypeError):
                max_age = None

        if max_age is not None or preferred_policies or preferred_level_types:
            pape_request = pape.Request(
                max_auth_age=max_age,
                preferred_auth_policies=preferred_policies,
                preferred_auth_level_types=preferred_level_types,
            )
            request.addExtension(pape_request)
        return request

    def get_consumer_store(self) -> OpenIdStore | None:
        return self.strategy.openid_store()

    def consumer(self):
        """Create an OpenID Consumer object for the given Django request."""
        if self._consumer is None:
            self._consumer = self.create_consumer(self.get_consumer_store())
        return self._consumer

    def create_consumer(self, store=None):
        return Consumer(self.strategy.openid_session_dict(SESSION_NAME), store)

    def uses_redirect(self):
        """Return true if openid request will be handled with redirect or
        HTML content will be returned.
        """
        return self.openid_request().shouldSendRedirect()

    def openid_request(self, params: dict[str, str] | None = None):
        """Return openid request"""
        try:
            return self.consumer().begin(url_add_parameters(self.openid_url(), params))
        except DiscoveryFailure as err:
            raise AuthException(self, f"OpenID discovery error: {err}") from err

    def openid_url(self):
        """Return service provider URL.
        This base class is generic accepting a POST parameter that specifies
        provider URL."""
        if self.URL:
            return self.URL
        if OPENID_ID_FIELD in self.data:
            return self.data[OPENID_ID_FIELD]
        raise AuthMissingParameter(self, OPENID_ID_FIELD)



from __future__ import annotations

import base64
import datetime
from calendar import timegm
from json import loads
from typing import TYPE_CHECKING, Any, Literal, cast

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from jwt import (
    ExpiredSignatureError,
    InvalidAudienceError,
    InvalidTokenError,
    PyJWTError,
)
from jwt.utils import base64url_decode

from social_core.backends.oauth import BaseOAuth2
from social_core.exceptions import (
    AuthInvalidParameter,
    AuthMissingParameter,
    AuthTokenError,
)
from social_core.utils import cache

if TYPE_CHECKING:
    from collections.abc import Mapping

    from jwt.types import Options
    from requests.auth import AuthBase

    from social_core.strategy import BaseStrategy


class OpenIdConnectAssociation:
    """Use Association model to save the nonce by force."""

    def __init__(self, handle, secret="", issued=0, lifetime=0, assoc_type="") -> None:
        self.handle = handle  # as nonce
        self.secret = secret.encode()  # not use
        self.issued = issued  # not use
        self.lifetime = lifetime  # not use
        self.assoc_type = assoc_type  # as state


class OpenIdConnectAuth(BaseOAuth2):
    """
    Base class for Open ID Connect backends.
    Currently only the code response type is supported.

    It can also be directly instantiated as a generic OIDC backend.
    To use it you will need to set at minimum:

    SOCIAL_AUTH_OIDC_OIDC_ENDPOINT = 'https://.....'  # endpoint without /.well-known/openid-configuration
    SOCIAL_AUTH_OIDC_KEY = '<client_id>'
    SOCIAL_AUTH_OIDC_SECRET = '<client_secret>'
    """

    name = "oidc"
    # Override OIDC_ENDPOINT in your subclass to enable autoconfig of OIDC
    OIDC_ENDPOINT: str | None = None
    ID_TOKEN_MAX_AGE = 600
    DEFAULT_SCOPE = ["openid", "profile", "email"]
    EXTRA_DATA = ["id_token", "refresh_token", ("sub", "id")]
    REDIRECT_STATE = False
    REVOKE_TOKEN_METHOD = "GET"
    ID_KEY = "sub"
    USERNAME_KEY = "preferred_username"
    JWT_ALGORITHMS = ["RS256"]
    JWT_DECODE_OPTIONS: Options = {}
    JWT_LEEWAY: float = 1.0  # seconds
    VALIDATE_AT_HASH: bool = True
    CUSTOM_AT_HASH_ALGO: str | None = None
    # When these options are unspecified, server will choose via openid autoconfiguration
    ID_TOKEN_ISSUER = ""
    ACCESS_TOKEN_URL = ""
    AUTHORIZATION_URL = ""
    REVOKE_TOKEN_URL = ""
    USERINFO_URL = ""
    JWKS_URI = ""
    TOKEN_ENDPOINT_AUTH_METHOD = ""
    # Optional parameters for Authentication Request
    DISPLAY = None
    PROMPT = None
    MAX_AGE = None
    UI_LOCALES = None
    ID_TOKEN_HINT = None
    LOGIN_HINT = None
    ACR_VALUES = None

    def __init__(
        self, strategy: BaseStrategy | None = None, redirect_uri: str | None = None
    ) -> None:
        super().__init__(strategy, redirect_uri=redirect_uri)
        self.id_token = None

    def get_setting_config(
        self, setting_name: str, oidc_name: str, default: str
    ) -> str:
        value = self.setting(setting_name, default)
        if not value:
            value = self.oidc_config().get(oidc_name)

        if not isinstance(value, str):
            raise AuthMissingParameter(self, setting_name)
        return value

    def authorization_url(self) -> str:
        return self.get_setting_config(
            "AUTHORIZATION_URL", "authorization_endpoint", self.AUTHORIZATION_URL
        )

    def access_token_url(self) -> str:
        return self.get_setting_config(
            "ACCESS_TOKEN_URL", "token_endpoint", self.ACCESS_TOKEN_URL
        )

    def revoke_token_url(self, token, uid) -> str:
        return self.get_setting_config(
            "REVOKE_TOKEN_URL", "revocation_endpoint", self.REVOKE_TOKEN_URL
        )

    def id_token_issuer(self) -> str:
        return self.get_setting_config(
            "ID_TOKEN_ISSUER", "issuer", self.ID_TOKEN_ISSUER
        )

    def userinfo_url(self) -> str:
        return self.get_setting_config(
            "USERINFO_URL", "userinfo_endpoint", self.USERINFO_URL
        )

    def jwks_uri(self) -> str:
        return self.get_setting_config("JWKS_URI", "jwks_uri", self.JWKS_URI)

    def use_basic_auth(self) -> bool:
        method = self.setting(
            "TOKEN_ENDPOINT_AUTH_METHOD", self.TOKEN_ENDPOINT_AUTH_METHOD
        )
        if method:
            return method == "client_secret_basic"
        methods = self.oidc_config().get("token_endpoint_auth_methods_supported", [])
        return not methods or "client_secret_basic" in methods

    def oidc_endpoint(self) -> str:
        return cast("str", self.setting("OIDC_ENDPOINT", self.OIDC_ENDPOINT))

    @cache(ttl=86400)
    def oidc_config(self) -> dict[Any, Any]:
        return self.get_json(f"{self.oidc_endpoint()}/.well-known/openid-configuration")

    @cache(ttl=86400)
    def get_jwks_keys(self):
        return self.get_remote_jwks_keys()

        # Add client secret as oct key so it can be used for HMAC signatures
        # client_id, client_secret = self.get_key_and_secret()
        # keys.append({'key': client_secret, 'kty': 'oct'})

    def get_remote_jwks_keys(self):
        response = self.request(self.jwks_uri())
        return loads(response.text)["keys"]

    def auth_params(self, state=None):  # noqa: C901, PLR0912
        """Return extra arguments needed on auth process."""
        params = super().auth_params(state)
        params["nonce"] = self.get_and_store_nonce(self.authorization_url(), state)

        display = self.setting("DISPLAY", default=self.DISPLAY)
        if display is not None:
            if not display:
                raise AuthMissingParameter(
                    self, "OpenID Connect display value cannot be empty string."
                )

            if display not in ("page", "popup", "touch", "wap"):
                raise AuthMissingParameter(
                    self, f"Invalid OpenID Connect display value: {display}"
                )

            params["display"] = display

        prompt = self.setting("PROMPT", default=self.PROMPT)
        if prompt is not None:
            if not prompt:
                raise AuthInvalidParameter(self, "prompt")

            for prompt_token in prompt.split():
                if prompt_token not in ("none", "login", "consent", "select_account"):
                    raise AuthInvalidParameter(self, "prompt")

            params["prompt"] = prompt

        max_age = self.setting("MAX_AGE", default=self.MAX_AGE)
        if max_age is not None:
            if max_age < 0:
                raise AuthInvalidParameter(self, "max_age")

            params["max_age"] = max_age

        ui_locales = self.setting("UI_LOCALES", default=self.UI_LOCALES)
        if ui_locales is not None:
            if not ui_locales:
                raise AuthInvalidParameter(self, "ui_locales")

            params["ui_locales"] = ui_locales

        id_token_hint = self.setting("ID_TOKEN_HINT", default=self.ID_TOKEN_HINT)
        if id_token_hint is not None:
            if not id_token_hint:
                raise AuthInvalidParameter(self, "id_token_hint")

            params["id_token_hint"] = id_token_hint

        login_hint = self.setting("LOGIN_HINT", default=self.LOGIN_HINT)
        if login_hint is not None:
            if not login_hint:
                raise AuthInvalidParameter(self, "login_hint")

            params["login_hint"] = login_hint

        acr_values = self.setting("ACR_VALUES", default=self.ACR_VALUES)
        if acr_values is not None:
            if not acr_values:
                raise AuthInvalidParameter(self, "acr_values")

            params["acr_values"] = acr_values

        return params

    def get_and_store_nonce(self, url, state):
        # Create a nonce
        nonce = self.strategy.random_string(64)
        # Store the nonce
        association = OpenIdConnectAssociation(nonce, assoc_type=state)
        self.strategy.storage.association.store(url, association)
        return nonce

    def get_nonce(self, nonce):
        try:
            return self.strategy.storage.association.get(
                server_url=self.authorization_url(), handle=nonce
            )[0]
        except IndexError:
            return None

    def remove_nonce(self, nonce_id) -> None:
        self.strategy.storage.association.remove([nonce_id])

    def validate_claims(self, id_token) -> None:
        utc_timestamp = timegm(datetime.datetime.now(datetime.timezone.utc).timetuple())

        if "nbf" in id_token and utc_timestamp < id_token["nbf"]:
            raise AuthTokenError(self, "Incorrect id_token: nbf")

        # Verify the token was issued in the last 10 minutes
        iat_leeway = self.setting("ID_TOKEN_MAX_AGE", self.ID_TOKEN_MAX_AGE)
        if utc_timestamp > id_token["iat"] + iat_leeway:
            raise AuthTokenError(self, "Incorrect id_token: iat")

        # Validate the nonce to ensure the request was not modified
        nonce = id_token.get("nonce")
        if not nonce:
            raise AuthTokenError(self, "Incorrect id_token: nonce")

        nonce_obj = self.get_nonce(nonce)
        if nonce_obj:
            self.remove_nonce(nonce_obj.id)
        else:
            raise AuthTokenError(self, "Incorrect id_token: nonce")

    def find_valid_key(self, id_token):
        kid = jwt.get_unverified_header(id_token).get("kid")

        keys = self.get_jwks_keys()
        if kid is not None:
            for key in keys:
                if kid == key.get("kid"):
                    break
            else:
                # In case the key id is not found in the cached keys, just
                # reload the JWKS keys. Ideally this should be done by
                # invalidating the cache.
                self.get_jwks_keys.invalidate()  # pyright: ignore[reportAttributeAccessIssue]
                keys = self.get_jwks_keys()

        for key in keys:
            if kid is None or kid == key.get("kid"):
                if "alg" not in key:
                    key["alg"] = cast(
                        "list[str]", self.setting("JWT_ALGORITHMS", self.JWT_ALGORITHMS)
                    )[0]
                rsakey = jwt.PyJWK(key)
                message, encoded_sig = id_token.rsplit(".", 1)
                decoded_sig = base64url_decode(encoded_sig.encode("utf-8"))
                if rsakey.Algorithm.verify(
                    message.encode("utf-8"), rsakey.key, decoded_sig
                ):
                    return key
        return None

    def validate_and_return_id_token(self, id_token, access_token):
        """
        Validates the id_token according to the steps at
        http://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation.
        """
        client_id, _client_secret = self.get_key_and_secret()

        key = self.find_valid_key(id_token)

        if not key:
            raise AuthTokenError(self, "Signature verification failed")

        rsakey = jwt.PyJWK(key)

        try:
            claims = jwt.decode(
                id_token,
                rsakey.key,
                algorithms=self.setting("JWT_ALGORITHMS", self.JWT_ALGORITHMS),
                audience=client_id,
                issuer=self.id_token_issuer(),
                options=self.setting("JWT_DECODE_OPTIONS", self.JWT_DECODE_OPTIONS),
                leeway=cast("int", self.setting("JWT_LEEWAY", self.JWT_LEEWAY)),
            )
        except ExpiredSignatureError as error:
            raise AuthTokenError(self, "Signature has expired") from error
        except InvalidAudienceError as error:
            # compatibility with jose error message
            raise AuthTokenError(self, "Token error: Invalid audience") from error
        except InvalidTokenError as error:
            raise AuthTokenError(self, str(error)) from error
        except PyJWTError as error:
            raise AuthTokenError(self, "Invalid signature") from error

        # pyjwt does not validate OIDC claims
        # see https://github.com/jpadilla/pyjwt/pull/296
        if not self.validate_at_hash(claims, access_token, key):
            raise AuthTokenError(self, "Invalid access token")

        self.validate_claims(claims)

        return claims

    def request_access_token(
        self,
        url: str,
        method: Literal["GET", "POST", "DELETE"] = "GET",
        headers: Mapping[str, str | bytes] | None = None,
        data: dict | None = None,
        json: dict | None = None,
        auth: tuple[str, str] | AuthBase | None = None,
        params: dict | None = None,
    ) -> dict[Any, Any]:
        """
        Retrieve the access token. Also, validate the id_token and
        store it (temporarily).
        """
        response = super().request_access_token(
            url,
            method=method,
            headers=headers,
            data=data,
            json=json,
            auth=auth,
            params=params,
        )
        self.id_token = self.validate_and_return_id_token(
            response["id_token"], response["access_token"]
        )
        return response

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            self.userinfo_url(), headers={"Authorization": f"Bearer {access_token}"}
        )

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", self.USERNAME_KEY)

        def get_value(key):
            if key in response:
                return response.get(key)
            if self.id_token is not None:
                return self.id_token.get(key)
            return None

        return {
            "username": get_value(username_key),
            "email": get_value("email"),
            "fullname": get_value("name"),
            "first_name": get_value("given_name"),
            "last_name": get_value("family_name"),
        }

    def validate_at_hash(self, claims, access_token, key):
        """
        Validate the 'at_hash' claim according to OpenID Connect specs.

        See: https://openid.net/specs/openid-connect-core-1_0.html#CodeIDToken
        """

        if not self.VALIDATE_AT_HASH:
            return True
        if "at_hash" not in claims:
            return True

        expected_hash = claims["at_hash"]
        calculated_hash = self.calc_at_hash(
            access_token, key["alg"], self.CUSTOM_AT_HASH_ALGO
        )
        return expected_hash == calculated_hash

    @staticmethod
    def calc_at_hash(access_token, algorithm, custom_at_hash_algo: str | None = None):
        """
        Calculates "at_hash" claim which is not done by pyjwt.
        Custom "at_hash" algorithm is used for non-standard token.

        See https://pyjwt.readthedocs.io/en/stable/usage.html#oidc-login-flow
        See https://github.com/python-social-auth/social-core/issues/1306
        """

        if not custom_at_hash_algo:
            alg_obj = jwt.get_algorithm_by_name(algorithm)
            digest = alg_obj.compute_hash_digest(access_token.encode("utf-8"))
            return (
                base64.urlsafe_b64encode(digest[: (len(digest) // 2)])
                .decode("utf-8")
                .rstrip("=")
            )

        algo_class_name = custom_at_hash_algo.upper()
        algo_class = getattr(hashes, algo_class_name, None)
        if algo_class is None:
            raise NotImplementedError(
                f"Unsupported custom at hash algorithm: {custom_at_hash_algo}"
            )

        hasher = hashes.Hash(algo_class(), backend=default_backend())
        hasher.update(access_token.encode("utf-8"))
        digest = hasher.finalize()
        half = digest[: (len(digest) // 2)]
        return base64.urlsafe_b64encode(half).decode("utf-8").rstrip("=")



from typing import Any

from social_core.backends.oauth import BaseOAuth2


class GrafanaOAuth2(BaseOAuth2):
    """Grafana OAuth authentication backend"""

    name = "grafana"
    AUTHORIZATION_URL = "https://grafana.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://grafana.com/api/oauth2/token"
    DEFAULT_SCOPE = ["profile", "email"]
    SCOPE_SEPARATOR = ","
    USER_DETAILS_URL = "https://grafana.com/api/oauth2/user"

    def get_user_details(self, response):
        """Return user details from Grafana account"""
        return {
            "username": response.get("login"),
            "email": response.get("email") or "",
            "first_name": response.get("name"),
            "last_name": "-",
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            self.USER_DETAILS_URL,
            headers={"Authorization": f"Bearer {access_token}"},
        )



from __future__ import annotations

import base64
import time
from typing import TYPE_CHECKING, Any, Literal, cast

import requests

from social_core.exceptions import AuthConnectionError, AuthUnknownError
from social_core.registry import REGISTRY
from social_core.utils import module_member, parse_qs, social_logger, user_agent

if TYPE_CHECKING:
    from collections.abc import Mapping

    from requests import Response
    from requests.auth import AuthBase

    from social_core.storage import PartialMixin, UserProtocol
    from social_core.strategy import BaseStrategy, HttpResponseProtocol


class BaseAuth:
    """A authentication backend that authenticates the user based on
    the provider response"""

    name = ""  # provider name, it's stored in database
    supports_inactive_user = False  # Django auth
    ID_KEY: str = ""
    EXTRA_DATA: list[str | tuple[str, str] | tuple[str, str, bool]] | None = None
    GET_ALL_EXTRA_DATA = False
    REQUIRES_EMAIL_VALIDATION = False
    SEND_USER_AGENT = True

    def __init__(
        self, strategy: BaseStrategy | None = None, redirect_uri: str | None = None
    ) -> None:
        self.strategy: BaseStrategy = (
            strategy if strategy is not None else REGISTRY.default_strategy
        )
        self.redirect_uri = redirect_uri
        self.data = self.strategy.request_data()
        self.redirect_uri = self.strategy.absolute_uri(self.redirect_uri)

    def log_debug(self, message, *args) -> None:
        social_logger.debug(f"{self.name}: {message}", *args)

    def log_warning(self, message, *args) -> None:
        social_logger.warning(f"{self.name}: {message}", *args)

    def setting(self, name: str, default=None):
        """Return setting value from strategy"""
        return self.strategy.setting(name, default=default, backend=self)

    def start(self) -> HttpResponseProtocol:
        if self.uses_redirect():
            return self.strategy.redirect(self.auth_url())
        return self.strategy.html(self.auth_html())

    def complete(self, *args, **kwargs) -> HttpResponseProtocol | UserProtocol | None:
        return self.auth_complete(*args, **kwargs)

    def auth_url(self) -> str:
        """Must return redirect URL to auth provider"""
        raise NotImplementedError("Implement in subclass")

    def auth_html(self) -> str:
        """Must return login HTML content returned by provider"""
        return "Implement in subclass"

    def auth_complete(
        self, *args, **kwargs
    ) -> HttpResponseProtocol | UserProtocol | None:
        """Completes login process, must return user instance"""
        raise NotImplementedError("Implement in subclass")

    def process_error(self, data) -> None:
        """Process data for errors, raise exception if needed.
        Call this method on any override of auth_complete."""

    def authenticate(
        self, *args, **kwargs
    ) -> UserProtocol | HttpResponseProtocol | None:
        """Authenticate user using social credentials

        Authentication is made if this is the correct backend, backend
        verification is made by kwargs inspection for current backend
        name presence.
        """
        # Validate backend and arguments. Require that the Social Auth
        # response be passed in as a keyword argument, to make sure we
        # don't match the username/password calling conventions of
        # authenticate.
        if (
            "backend" not in kwargs
            or kwargs["backend"].name != self.name
            or "strategy" not in kwargs
            or "response" not in kwargs
        ):
            return None

        self.strategy = kwargs.get("strategy") or self.strategy
        self.redirect_uri = kwargs.get("redirect_uri") or self.redirect_uri
        self.data = self.strategy.request_data()
        kwargs.setdefault("is_new", False)
        pipeline = self.strategy.get_pipeline(self)
        args, kwargs = self.strategy.clean_authenticate_args(*args, **kwargs)
        return self.pipeline(pipeline, *args, **kwargs)

    def pipeline(
        self, pipeline, pipeline_index: int = 0, *args, **kwargs
    ) -> UserProtocol | HttpResponseProtocol | None:
        out = self.run_pipeline(pipeline, pipeline_index, *args, **kwargs)
        if not isinstance(out, dict):
            return cast("HttpResponseProtocol", out)
        user = cast("UserProtocol | None", out.get("user"))
        if user:
            user.social_user = out.get("social")  # type: ignore[attr-defined]
            user.is_new = out.get("is_new")  # type: ignore[attr-defined]
        return user

    def disconnect(self, *args, **kwargs) -> dict:
        pipeline = self.strategy.get_disconnect_pipeline(self)
        kwargs["name"] = self.name
        kwargs["user_storage"] = self.strategy.storage.user
        return self.run_pipeline(pipeline, *args, **kwargs)

    def run_pipeline(
        self, pipeline: list[str], pipeline_index=0, *args, **kwargs
    ) -> dict:
        out = kwargs.copy()
        out.setdefault("strategy", self.strategy)
        out.setdefault("backend", out.pop(self.name, None) or self)
        out.setdefault("request", self.strategy.request_data())
        out.setdefault("details", {})

        if (
            not isinstance(pipeline_index, int)
            or pipeline_index < 0
            or pipeline_index >= len(pipeline)
        ):
            pipeline_index = 0

        for idx, name in enumerate(pipeline[pipeline_index:]):
            out["pipeline_index"] = pipeline_index + idx
            func = module_member(name)
            result = func(*args, **out) or {}
            if not isinstance(result, dict):
                return result
            out.update(result)
        return out

    def extra_data(
        self,
        user: UserProtocol | None,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Return default extra data to store in extra_data field"""
        data: dict[str, Any] = {
            # store the last time authentication took place
            "auth_time": int(time.time())
        }
        extra_data_entries: (
            list[str] | list[str | tuple[str, str] | tuple[str, str, bool]]
        ) = []
        if self.GET_ALL_EXTRA_DATA or self.setting("GET_ALL_EXTRA_DATA", False):
            extra_data_entries = list(response.keys())
        else:
            extra_data_entries = (self.EXTRA_DATA or []) + cast(
                "list[str | tuple[str, str] | tuple[str, str, bool]]",
                self.setting("EXTRA_DATA", []),
            )
        for entry in extra_data_entries:
            if isinstance(entry, list):
                entry = tuple(cast("list[str]", entry))
            discard = False
            if isinstance(entry, str):
                name = alias = entry
            elif len(entry) == 3:
                name, alias, discard = entry
            elif len(entry) == 2:
                name, alias = entry  # ty: ignore[invalid-assignment]
            elif len(entry) == 1:
                name = alias = entry[0]
            else:
                raise AuthUnknownError(self, f"Invalid EXTRA_DATA item: {entry!r}")
            value = response.get(name, details.get(name, details.get(alias)))
            if discard and not value:
                continue
            data[alias] = value
        return data

    def auth_allowed(self, response, details):
        """Return True if the user should be allowed to authenticate, by
        default check if email is whitelisted (if there's a whitelist)"""
        emails = [
            email.lower()
            for email in cast("list[str]", self.setting("WHITELISTED_EMAILS", []))
        ]
        domains = [
            domain.lower()
            for domain in cast("list[str]", self.setting("WHITELISTED_DOMAINS", []))
        ]
        email = details.get("email")
        allowed = True
        if email and (emails or domains):
            email = email.lower()
            domain = email.split("@", 1)[1]
            allowed = email in emails or domain in domains
        return allowed

    def id_key(self) -> str:
        """Return the ID_KEY to use for this backend, checking settings first."""
        return self.setting("ID_KEY") or self.ID_KEY

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response or details."""
        id_key = self.id_key()
        if details:
            user_id = details.get(id_key)
            if user_id:
                return user_id
        return response.get(id_key)

    def get_user_details(self, response) -> dict[str, Any]:
        """Must return user details in a know internal struct:
        {'username': <username if any>,
         'email': <user email if any>,
         'fullname': <user full name if any>,
         'first_name': <user first name if any>,
         'last_name': <user last name if any>}
        """
        raise NotImplementedError("Implement in subclass")

    def get_user_names(self, fullname="", first_name="", last_name=""):
        # Avoid None values
        fullname = fullname or ""
        first_name = first_name or ""
        last_name = last_name or ""
        if fullname and not (first_name or last_name):
            try:
                first_name, last_name = fullname.split(" ", 1)
            except ValueError:
                first_name = first_name or fullname or ""
                last_name = last_name or ""
        fullname = fullname or f"{first_name} {last_name}"
        return fullname.strip(), first_name.strip(), last_name.strip()

    def get_user(self, user_id):
        """
        Return user with given ID from the User model used by this backend.
        This is called by django.contrib.auth.middleware.
        """
        return self.strategy.get_user(user_id)

    def continue_pipeline(
        self, partial: PartialMixin
    ) -> UserProtocol | HttpResponseProtocol | None:
        """Continue previous halted pipeline"""
        return self.strategy.authenticate(
            self, *partial.args, pipeline_index=partial.next_step, **partial.kwargs
        )

    def auth_extra_arguments(self) -> dict[str, str]:
        """Return extra arguments needed on auth process. The defaults can be
        overridden by GET parameters."""
        extra_arguments = self.setting("AUTH_EXTRA_ARGUMENTS", {}).copy()
        extra_arguments.update(
            (key, self.data[key]) for key in extra_arguments if key in self.data
        )
        return extra_arguments

    def uses_redirect(self) -> bool:
        """Return True if this provider uses redirect url method,
        otherwise return false."""
        return True

    def request(  # noqa: PLR0913
        self,
        url: str,
        *,
        method: Literal["GET", "POST", "DELETE"] = "GET",
        headers: Mapping[str, str | bytes] | None = None,
        data: dict | None = None,
        json: dict | None = None,
        auth: tuple[str, str] | AuthBase | None = None,
        params: dict | None = None,
        timeout: float | None = None,
    ) -> Response:
        headers = {} if headers is None else dict(headers)
        proxies = self.setting("PROXIES")
        verify = self.setting("VERIFY_SSL", True)

        if timeout is None:
            timeout = self.setting("REQUESTS_TIMEOUT") or self.setting(
                "URLOPEN_TIMEOUT"
            )

        if self.SEND_USER_AGENT and "User-Agent" not in headers:
            headers["User-Agent"] = self.setting("USER_AGENT") or user_agent()

        try:
            response = requests.request(
                method,
                url,
                headers=headers,
                data=data,
                json=json,
                auth=auth,
                params=params,
                timeout=timeout,
                proxies=proxies,
                verify=verify,
            )
        except requests.ConnectionError as err:
            raise AuthConnectionError(self, str(err)) from err
        response.raise_for_status()
        return response

    def get_json(  # noqa: PLR0913
        self,
        url: str,
        method: Literal["GET", "POST", "DELETE"] = "GET",
        headers: Mapping[str, str | bytes] | None = None,
        data: dict | None = None,
        json: dict | None = None,
        auth: tuple[str, str] | AuthBase | None = None,
        params: dict | None = None,
        timeout: float | None = None,
    ) -> dict[Any, Any]:
        return self.request(
            url,
            method=method,
            headers=headers,
            data=data,
            json=json,
            auth=auth,
            params=params,
            timeout=timeout,
        ).json()

    def get_querystring(self, url, *args, **kwargs) -> dict[str, str]:
        return parse_qs(self.request(url, *args, **kwargs).text)

    def get_key_and_secret(self) -> tuple[str, str]:
        """Return tuple with Consumer Key and Consumer Secret for current
        service provider. Must return (key, secret), order *must* be respected.
        """
        return cast("str", self.setting("KEY")), cast("str", self.setting("SECRET"))

    def get_key_and_secret_basic_auth(self) -> bytes:
        """Generate HTTP Basic Authentication header value from KEY and SECRET.

        Returns:
            Basic authentication value in the format b"Basic <base64-encoded-credentials>"
        """
        key, secret = self.get_key_and_secret()
        credentials = f"{key}:{secret}".encode()
        encoded = base64.b64encode(credentials)
        return b"Basic " + encoded



"""Models mixins for Social Auth"""

from __future__ import annotations

import base64
import re
import uuid
from abc import abstractmethod
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Protocol

from openid.association import Association as OpenIdAssociation

from .exceptions import InvalidExpiryValue, MissingBackend

if TYPE_CHECKING:
    from collections.abc import Callable

    from social_core.backends.base import BaseAuth
    from social_core.strategy import BaseStrategy

NO_ASCII_REGEX = re.compile(r"[^\x00-\x7F]+")
NO_SPECIAL_REGEX = re.compile(r"[^\w.@+_-]+", re.UNICODE)


class UserProtocol(Protocol):
    id: int
    username: str
    is_active: bool | Callable[[], bool]
    is_authenticated: bool | Callable[[], bool]

    # Set in BaseAuth.pipeline
    # social_user: UserMixin
    # is_new: bool


class UserMixin:
    # Consider tokens that expire in 5 seconds as already expired
    ACCESS_TOKEN_EXPIRED_THRESHOLD = 5

    provider = ""
    uid: str
    user: UserProtocol
    extra_data: dict[str, Any]

    @abstractmethod
    def save(self): ...

    def get_backend(self, strategy: BaseStrategy) -> type[BaseAuth]:
        return strategy.get_backend_class(self.provider)

    def get_backend_instance(self, strategy: BaseStrategy) -> BaseAuth | None:
        try:
            return strategy.get_backend(self.provider)
        except MissingBackend:
            return None

    @property
    def access_token(self) -> str | None:
        """Return access_token stored in extra_data or None"""
        return self.extra_data.get("access_token")

    def refresh_token(self, strategy: BaseStrategy, *args, **kwargs) -> None:
        token = self.extra_data.get("refresh_token") or self.extra_data.get(
            "access_token"
        )
        backend = self.get_backend_instance(strategy)
        if token and backend and hasattr(backend, "refresh_token"):
            response = backend.refresh_token(token, *args, **kwargs)  # type: ignore[attr-defined]
            extra_data = backend.extra_data(
                self.user, self.uid, response, self.extra_data or {}, {}
            )
            if self.set_extra_data(extra_data):
                self.save()

    def _compute_expiration_from_timestamp(
        self, value: int | str, field_name: str = "expires"
    ) -> timedelta:
        """Compute expiration timedelta from an absolute timestamp."""
        try:
            timestamp = int(value)
        except (ValueError, TypeError) as e:
            raise InvalidExpiryValue(field_name, value) from e

        try:
            now = datetime.now(timezone.utc)
            expiry_time = datetime.fromtimestamp(timestamp, tz=timezone.utc)
            return expiry_time - now
        except (OSError, ValueError) as e:
            raise InvalidExpiryValue(field_name, value) from e

    def _compute_expiration_from_relative(
        self, value: int | str, field_name: str = "expires"
    ) -> timedelta:
        """Compute expiration timedelta from relative seconds."""
        try:
            seconds = int(value)
        except (ValueError, TypeError) as e:
            raise InvalidExpiryValue(field_name, value) from e

        auth_time = self.extra_data.get("auth_time")
        if auth_time:
            try:
                auth_timestamp = int(auth_time)
            except (ValueError, TypeError):
                # Invalid auth_time value, fall back to treating as seconds from now
                pass
            else:
                try:
                    now = datetime.now(timezone.utc)
                    reference = datetime.fromtimestamp(auth_timestamp, tz=timezone.utc)
                    return (reference + timedelta(seconds=seconds)) - now
                except (OSError, ValueError):
                    # auth_time timestamp out of range, fall back to treating as seconds from now
                    pass
        # If no auth_time or invalid auth_time, treat as seconds from now
        return timedelta(seconds=seconds)

    def expiration_timedelta(self) -> timedelta | None:
        """Return provider session live seconds.

        Returns a timedelta ready to use with session.set_expiry().
        If provider returns a timestamp instead of session seconds to live, the
        timedelta is inferred from current time (using UTC timezone).

        Handles three types of expiration data:
        - expires_on: Always treated as absolute timestamp
        - expires_in: Always treated as relative seconds from auth_time
        - expires: Uses heuristic (>63072000 = 2 years) to distinguish timestamp vs relative
        """
        if not self.extra_data:
            return None

        # Check for expires_on (absolute timestamp)
        expires_on = self.extra_data.get("expires_on")
        if expires_on is not None:
            return self._compute_expiration_from_timestamp(expires_on, "expires_on")

        # Check for expires_in (relative seconds from auth_time)
        expires_in = self.extra_data.get("expires_in")
        if expires_in is not None:
            return self._compute_expiration_from_relative(expires_in, "expires_in")

        # Check for expires (use heuristic to determine type)
        return self._handle_expires_field()

    def _handle_expires_field(self) -> timedelta | None:
        """Handle the generic expires field using heuristic."""
        expires = self.extra_data.get("expires")
        if expires is None:
            return None

        try:
            expires_int = int(expires)
        except (ValueError, TypeError) as e:
            raise InvalidExpiryValue("expires", expires) from e

        # Use 2 years (63072000 seconds) as threshold to distinguish
        # absolute timestamps from relative seconds
        # Most tokens expire in hours/days/months, timestamps are much larger
        TIMESTAMP_THRESHOLD = 63072000

        if expires_int > TIMESTAMP_THRESHOLD:
            # Likely an absolute timestamp, try treating as expires_on
            return self._compute_expiration_from_timestamp(expires_int, "expires")

        # Treat as relative seconds (like expires_in)
        return self._compute_expiration_from_relative(expires_int, "expires")

    def expiration_datetime(self):
        # backward compatible alias
        return self.expiration_timedelta()

    def access_token_expired(self):
        """Return true / false if access token is already expired"""
        expiration = self.expiration_timedelta()
        return (
            expiration
            and expiration.total_seconds() <= self.ACCESS_TOKEN_EXPIRED_THRESHOLD
        )

    def get_access_token(self, strategy: BaseStrategy) -> str | None:
        """Returns a valid access token."""
        if self.access_token_expired():
            self.refresh_token(strategy)
        return self.access_token

    def set_extra_data(self, extra_data: dict[str, Any] | None = None) -> bool:
        if extra_data and self.extra_data != extra_data:
            if self.extra_data and not isinstance(self.extra_data, str):
                self.extra_data.update(extra_data)
            else:
                self.extra_data = extra_data
            return True
        return False

    @classmethod
    def clean_username(cls, value: str) -> str:
        """Clean username removing any unsupported character"""
        value = NO_ASCII_REGEX.sub("", value)
        return NO_SPECIAL_REGEX.sub("", value)

    @classmethod
    def changed(cls, user: UserProtocol) -> None:
        """The given user instance is ready to be saved"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get_username(cls, user: UserProtocol) -> str:
        """Return the username for given user"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def user_model(cls) -> type[UserProtocol]:
        """Return the user model"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def username_max_length(cls) -> int:
        """Return the max length for username"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def allowed_to_disconnect(
        cls, user: UserProtocol, backend_name: str, association_id=None
    ) -> bool:
        """Return if it's safe to disconnect the social account for the
        given user"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def disconnect(cls, entry):
        """Disconnect the social account for the given user"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def user_exists(cls, *args, **kwargs) -> bool:
        """
        Return True/False if a User instance exists with the given arguments.
        Arguments are directly passed to filter() manager method.
        """
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def create_user(cls, *args, **kwargs):
        """Create a user instance"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get_user(cls, pk):
        """Return user instance for given id"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get_users_by_email(cls, email: str):
        """Return users instances for given email address"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get_social_auth(cls, provider: str, uid: int):
        """Return UserSocialAuth for given provider and uid"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get_social_auth_for_user(
        cls,
        user: UserProtocol,
        provider: str | None = None,
        # pylint: disable-next=redefined-builtin
        id: int | None = None,  # noqa: A002
    ):
        """Return all the UserSocialAuth instances for given user"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def create_social_auth(cls, user: UserProtocol, uid: int, provider: str):
        """Create a UserSocialAuth instance for given user"""
        raise NotImplementedError("Implement in subclass")


class NonceMixin:
    """One use numbers"""

    server_url = ""
    timestamp = 0
    salt = ""

    @classmethod
    def use(cls, server_url: str, timestamp, salt: str):
        """Create a Nonce instance"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get(cls, server_url: str, salt: str):
        """Retrieve a Nonce instance"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def delete(cls, nonce):
        """Delete a Nonce instance"""
        raise NotImplementedError("Implement in subclass")


class AssociationMixin:
    """OpenId account association"""

    server_url = ""
    handle = ""
    secret = ""
    issued = 0
    lifetime = 0
    assoc_type = ""

    @classmethod
    def oids(cls, server_url, handle=None):
        kwargs = {"server_url": server_url}
        if handle is not None:
            kwargs["handle"] = handle
        return sorted(
            ((assoc.id, cls.openid_association(assoc)) for assoc in cls.get(**kwargs)),
            key=lambda x: x[1].issued,
            reverse=True,
        )

    @classmethod
    def openid_association(cls, assoc):
        secret = assoc.secret
        if not isinstance(secret, bytes):
            secret = secret.encode()
        return OpenIdAssociation(
            assoc.handle,
            base64.decodebytes(secret),
            assoc.issued,
            assoc.lifetime,
            assoc.assoc_type,
        )

    @classmethod
    def store(cls, server_url, association):
        """Create an Association instance"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def get(cls, server_url: str | None = None, handle: str | None = None):
        """Get an Association instance"""
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def remove(cls, ids_to_delete):
        """Remove an Association instance"""
        raise NotImplementedError("Implement in subclass")


class CodeMixin:
    email = ""
    code = ""
    verified = False

    @abstractmethod
    def save(self): ...

    def verify(self) -> None:
        self.verified = True
        self.save()

    @classmethod
    def generate_code(cls):
        return uuid.uuid4().hex

    @classmethod
    def make_code(cls, email: str) -> CodeMixin:
        code = cls()
        code.email = email
        code.code = cls.generate_code()
        code.verified = False
        code.save()
        return code

    @classmethod
    def get_code(cls, code):
        raise NotImplementedError("Implement in subclass")


class PartialMixin:
    token = ""
    data: dict[str, Any] = {}
    next_step: int
    backend = ""

    @property
    def args(self):
        return self.data.get("args", [])

    @args.setter
    def args(self, value) -> None:
        self.data["args"] = value

    @abstractmethod
    def save(self): ...

    @property
    def kwargs(self):
        return self.data.get("kwargs", {})

    @kwargs.setter
    def kwargs(self, value) -> None:
        self.data["kwargs"] = value

    def extend_kwargs(self, values) -> None:
        self.data["kwargs"].update(values)

    @classmethod
    def generate_token(cls) -> str:
        return uuid.uuid4().hex

    @classmethod
    def load(cls, token: str) -> PartialMixin | None:
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def destroy(cls, token: str):
        raise NotImplementedError("Implement in subclass")

    @classmethod
    def prepare(
        cls, backend: str, next_step: int, data: dict[str, Any]
    ) -> PartialMixin:
        partial = cls()
        partial.backend = backend
        partial.next_step = next_step
        partial.data = data
        partial.token = cls.generate_token()
        return partial

    @classmethod
    def store(cls, partial: PartialMixin) -> PartialMixin:
        partial.save()
        return partial


class BaseStorage:
    user = UserMixin
    nonce = NonceMixin
    association = AssociationMixin
    code = CodeMixin
    partial = PartialMixin

    @classmethod
    def is_integrity_error(cls, exception) -> bool:
        """Check if given exception flags an integrity error in the DB"""
        raise NotImplementedError("Implement in subclass")



from urllib.parse import urlencode

from requests import HTTPError

from social_core.exceptions import AuthCanceled

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class EvernoteOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.evernote.EvernoteOAuth"
    expected_username = "101010"
    access_token_body = urlencode(
        {
            "edam_webApiUrlPrefix": "https://sandbox.evernote.com/shard/s1/",
            "edam_shard": "s1",
            "oauth_token": "foobar",
            "edam_expires": "1395118279645",
            "edam_userId": "101010",
            "edam_noteStoreUrl": "https://sandbox.evernote.com/shard/s1/notestore",
        }
    )
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class EvernoteOAuth1CanceledTest(EvernoteOAuth1Test):
    access_token_status = 401

    def test_login(self) -> None:
        with self.assertRaises(AuthCanceled) as cm:
            self.do_login()
        self.assertIsNotNone(cm.exception.response)

    def test_partial_pipeline(self) -> None:
        with self.assertRaises(AuthCanceled) as cm:
            self.do_partial_pipeline()
        self.assertIsNotNone(cm.exception.response)


class EvernoteOAuth1ErrorTest(EvernoteOAuth1Test):
    access_token_status = 500

    def test_login(self) -> None:
        with self.assertRaises(HTTPError):
            self.do_login()

    def test_partial_pipeline(self) -> None:
        with self.assertRaises(HTTPError):
            self.do_partial_pipeline()



"""
Yandex OpenID and OAuth2 support.

This contribution adds support for Yandex.ru OpenID service in the form
openid.yandex.ru/user. Username is retrieved from the identity url.

If username is not specified, OpenID 2.0 url used for authentication.
"""

from typing import Any
from urllib.parse import urlsplit

from .oauth import BaseOAuth2
from .open_id import OpenIdAuth


class YandexOpenId(OpenIdAuth):
    """Yandex OpenID authentication backend"""

    name = "yandex-openid"
    URL = "http://openid.yandex.ru"

    def get_user_id(self, details, response):
        return details["email"] or response.identity_url

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values["username"] = values.get("username") or urlsplit(
            response.identity_url
        ).path.strip("/")
        values["email"] = values.get("email", "")
        return values


class YandexOAuth2(BaseOAuth2):
    """Legacy Yandex OAuth2 authentication backend"""

    name = "yandex-oauth2"
    AUTHORIZATION_URL = "https://oauth.yandex.com/authorize"
    ACCESS_TOKEN_URL = "https://oauth.yandex.com/token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("real_name") or response.get("display_name") or ""
        )
        email = response.get("default_email")
        if not email:
            emails = response.get("emails")
            email = emails[0] if emails else ""
        return {
            "username": response.get("display_name"),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            "https://login.yandex.ru/info",
            params={"oauth_token": access_token, "format": "json"},
        )


class YaruOAuth2(BaseOAuth2):
    name = "yaru"
    AUTHORIZATION_URL = "https://oauth.yandex.com/authorize"
    ACCESS_TOKEN_URL = "https://oauth.yandex.com/token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("real_name") or response.get("display_name") or ""
        )
        email = response.get("default_email")
        if not email:
            emails = response.get("emails")
            email = emails[0] if emails else ""
        return {
            "username": response.get("display_name"),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            "https://login.yandex.ru/info",
            params={"oauth_token": access_token, "format": "json"},
        )



"""
LoginRadius BaseOAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/loginradius.html
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from .oauth import BaseOAuth2

if TYPE_CHECKING:
    from collections.abc import Mapping

    from requests.auth import AuthBase


class LoginRadiusAuth(BaseOAuth2):
    """LoginRadius BaseOAuth2 authentication backend."""

    name = "loginradius"
    ID_KEY = "ID"
    ACCESS_TOKEN_URL = "https://api.loginradius.com/api/v2/access_token"
    PROFILE_URL = "https://api.loginradius.com/api/v2/userprofile"
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def uses_redirect(self) -> bool:
        """Return False because we return HTML instead."""
        return False

    def auth_html(self):
        key, _secret = self.get_key_and_secret()
        tpl = self.setting("TEMPLATE", "loginradius.html")
        return self.strategy.render_html(
            tpl=tpl,
            context={
                "backend": self,
                "LOGINRADIUS_KEY": key,
                "LOGINRADIUS_REDIRECT_URL": self.get_redirect_uri(),
            },
        )

    def request_access_token(
        self,
        url: str,
        method: Literal["GET", "POST", "DELETE"] = "GET",
        headers: Mapping[str, str | bytes] | None = None,
        data: dict | None = None,
        json: dict | None = None,
        auth: tuple[str, str] | AuthBase | None = None,
        params: dict | None = None,
    ) -> dict[Any, Any]:
        return super().request_access_token(
            url,
            method=method,
            headers=headers,
            data=data,
            json=json,
            auth=auth,
            params={
                "token": self.data.get("token"),
                "secret": self.setting("SECRET"),
            },
        )

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service. Implement in subclass."""
        return self.get_json(
            self.PROFILE_URL,
            params={"access_token": access_token},
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method="GET",
        )

    def get_user_details(self, response):
        """Must return user details in a know internal struct:
        {'username': <username if any>,
         'email': <user email if any>,
         'fullname': <user full name if any>,
         'first_name': <user first name if any>,
         'last_name': <user last name if any>}
        """
        return {
            "username": response["NickName"] or "",
            "email": response["Email"][0]["Value"] or "",
            "fullname": response["FullName"] or "",
            "first_name": response["FirstName"] or "",
            "last_name": response["LastName"] or "",
        }

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response. Since LoginRadius handles multiple providers, we need to
        distinguish them to prevent conflicts."""
        return f"{response.get('Provider')}-{response.get(self.id_key())}"



# author:duoduo3369@gmail.com  https://github.com/duoduo369
"""
Weixin OAuth2 backend
"""

from typing import Any
from urllib.parse import urlencode

from requests import HTTPError

from social_core.exceptions import AuthCanceled, AuthUnknownError

from .oauth import BaseOAuth2


class WeixinOAuth2(BaseOAuth2):
    """Weixin OAuth authentication backend"""

    name = "weixin"
    ID_KEY = "openid"
    AUTHORIZATION_URL = "https://open.weixin.qq.com/connect/qrconnect"
    ACCESS_TOKEN_URL = "https://api.weixin.qq.com/sns/oauth2/access_token"
    DEFAULT_SCOPE = ["snsapi_login"]
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("nickname", "username"),
        ("headimgurl", "profile_image_url"),
    ]

    def get_user_details(self, response):
        """Return user details from Weixin. API URL is:
        https://api.weixin.qq.com/sns/userinfo
        """
        if self.setting("DOMAIN_AS_USERNAME"):
            username = response.get("domain", "")
        else:
            username = response.get("nickname", "")
        return {
            "username": username,
            "profile_image_url": response.get("headimgurl", ""),
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        data = self.get_json(
            "https://api.weixin.qq.com/sns/userinfo",
            params={
                "access_token": access_token,
                "openid": kwargs["response"]["openid"],
            },
        )
        nickname = data.get("nickname")
        if nickname:
            # weixin api has some encode bug, here need handle
            data["nickname"] = nickname.encode("raw_unicode_escape").decode("utf-8")
        return data

    def auth_params(self, state=None):
        appid, _secret = self.get_key_and_secret()
        params = {"appid": appid, "redirect_uri": self.get_redirect_uri(state)}
        if self.STATE_PARAMETER and state:
            params["state"] = state
        if self.RESPONSE_TYPE:
            params["response_type"] = self.RESPONSE_TYPE
        return params

    def auth_complete_params(self, state=None):
        appid, secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "appid": appid,
            "secret": secret,
            "redirect_uri": self.get_redirect_uri(state),
        }

    def refresh_token_params(self, token, *args, **kwargs):
        appid, secret = self.get_key_and_secret()
        return {
            "refresh_token": token,
            "grant_type": "refresh_token",
            "appid": appid,
            "secret": secret,
        }

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        try:
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(self.validate_state()),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD,
            )
        except HTTPError as err:
            if err.response.status_code == 400:
                raise AuthCanceled(self, response=err.response) from err
            raise
        except KeyError as err:
            raise AuthUnknownError(self) from err
        if "errcode" in response:
            raise AuthCanceled(self)
        self.process_error(response)
        return self.do_auth(
            response["access_token"], *args, response=response, **kwargs
        )


class WeixinOAuth2APP(WeixinOAuth2):
    """
    Weixin OAuth authentication backend

    Can't use in web, only in weixin app
    """

    name = "weixinapp"
    ID_KEY = "openid"
    AUTHORIZATION_URL = "https://open.weixin.qq.com/connect/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://api.weixin.qq.com/sns/oauth2/access_token"
    REDIRECT_STATE = False

    def auth_url(self) -> str:
        if self.STATE_PARAMETER or self.REDIRECT_STATE:
            # Store state in session for further request validation. The state
            # value is passed as state parameter (as specified in OAuth2 spec),
            # but also added to redirect, that way we can still verify the
            # request if the provider doesn't implement the state parameter.
            # Reuse token if any.
            name = f"{self.name}_state"
            state = self.strategy.session_get(name)
            if state is None:
                state = self.state_token()
                self.strategy.session_set(name, state)
        else:
            state = None

        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())
        params = urlencode(sorted(params.items()))
        return f"{self.authorization_url()}?{params}#wechat_redirect"

    def auth_complete_params(self, state=None):
        appid, secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "appid": appid,
            "secret": secret,
        }

    def validate_state(self) -> None:
        return None

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        try:
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(self.validate_state()),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD,
            )
        except HTTPError as err:
            if err.response.status_code == 400:
                raise AuthCanceled(self) from err
            raise
        except KeyError as err:
            raise AuthUnknownError(self) from err

        if "errcode" in response:
            raise AuthCanceled(self)
        self.process_error(response)
        return self.do_auth(
            response["access_token"], *args, response=response, **kwargs
        )



import unittest

from social_core.backends.base import BaseAuth
from social_core.exceptions import (
    AuthAlreadyAssociated,
    AuthCanceled,
    AuthException,
    AuthFailed,
    AuthForbidden,
    AuthMissingParameter,
    AuthStateForbidden,
    AuthStateMissing,
    AuthTokenError,
    AuthTokenRevoked,
    AuthUnknownError,
    AuthUnreachableProvider,
    InvalidEmail,
    MissingBackend,
    NotAllowedToDisconnect,
    SocialAuthBaseException,
    StrategyMissingBackendError,
    WrongBackend,
)

from .models import TestStorage
from .strategy import TestStrategy


class BaseExceptionTestCase(unittest.TestCase):
    exception: SocialAuthBaseException = SocialAuthBaseException("base test")
    expected_message: str = "base test"

    def test_exception_message(self) -> None:
        try:
            raise self.exception
        except SocialAuthBaseException as err:
            self.assertEqual(str(err), self.expected_message)


class WrongBackendTest(BaseExceptionTestCase):
    exception = WrongBackend("foobar")
    expected_message = 'Incorrect authentication service "foobar"'


class AuthFailedTest(BaseExceptionTestCase):
    exception = AuthFailed(BaseAuth(TestStrategy(TestStorage)), "wrong_user")
    expected_message = "Authentication failed: wrong_user"


class AuthFailedDeniedTest(BaseExceptionTestCase):
    exception = AuthFailed(BaseAuth(TestStrategy(TestStorage)), "access_denied")
    expected_message = "Authentication process was canceled"


class AuthTokenErrorTest(BaseExceptionTestCase):
    exception = AuthTokenError(BaseAuth(TestStrategy(TestStorage)), "Incorrect tokens")
    expected_message = "Token error: Incorrect tokens"


class AuthMissingParameterTest(BaseExceptionTestCase):
    exception = AuthMissingParameter(BaseAuth(TestStrategy(TestStorage)), "username")
    expected_message = "Missing needed parameter username"


class AuthStateMissingTest(BaseExceptionTestCase):
    exception = AuthStateMissing(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "Session value state missing."


class NotAllowedToDisconnectTest(BaseExceptionTestCase):
    exception = NotAllowedToDisconnect()
    expected_message = "This account is not allowed to be disconnected."


class AuthExceptionTest(BaseExceptionTestCase):
    exception = AuthException(BaseAuth(TestStrategy(TestStorage)), "message")
    expected_message = "message"


class AuthCanceledTest(BaseExceptionTestCase):
    exception = AuthCanceled(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "Authentication process canceled"


class AuthCanceledWithExtraMessageTest(BaseExceptionTestCase):
    exception = AuthCanceled(BaseAuth(TestStrategy(TestStorage)), "error_message")
    expected_message = "Authentication process canceled: error_message"


class AuthUnknownErrorTest(BaseExceptionTestCase):
    exception = AuthUnknownError(BaseAuth(TestStrategy(TestStorage)), "some error")
    expected_message = "An unknown error happened while authenticating some error"


class AuthStateForbiddenTest(BaseExceptionTestCase):
    exception = AuthStateForbidden(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "Wrong state parameter given."


class AuthAlreadyAssociatedTest(BaseExceptionTestCase):
    exception = AuthAlreadyAssociated(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "This account is already in use."


class AuthTokenRevokedTest(BaseExceptionTestCase):
    exception = AuthTokenRevoked(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "User revoke access to the token"


class AuthForbiddenTest(BaseExceptionTestCase):
    exception = AuthForbidden(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "Your credentials aren't allowed"


class AuthUnreachableProviderTest(BaseExceptionTestCase):
    exception = AuthUnreachableProvider(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "The authentication provider could not be reached"


class InvalidEmailTest(BaseExceptionTestCase):
    exception = InvalidEmail(BaseAuth(TestStrategy(TestStorage)))
    expected_message = "Email couldn't be validated"


class MissingBackendTest(BaseExceptionTestCase):
    exception = MissingBackend("backend")
    expected_message = 'Missing backend "backend" entry'


class StrategyMissingBackendErrorTest(BaseExceptionTestCase):
    exception = StrategyMissingBackendError()
    expected_message = "Strategy storage backend is not configured"



"""
Amazon OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/amazon.html
"""

from typing import Any

from .oauth import BaseOAuth2


class AmazonOAuth2(BaseOAuth2):
    name = "amazon"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://www.amazon.com/ap/oa"
    ACCESS_TOKEN_URL = "https://api.amazon.com/auth/o2/token"
    DEFAULT_SCOPE = ["profile"]
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("user_id", "user_id"),
        ("postal_code", "postal_code"),
    ]

    def get_user_details(self, response):
        """Return user details from amazon account"""
        name = response.get("name") or ""
        fullname, first_name, last_name = self.get_user_names(name)
        return {
            "username": name,
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Grab user profile information from amazon."""
        response = self.get_json(
            "https://api.amazon.com/user/profile", params={"access_token": access_token}
        )
        if "Profile" in response:
            response = {
                "user_id": response["Profile"]["CustomerId"],
                "name": response["Profile"]["Name"],
                "email": response["Profile"]["PrimaryEmail"],
            }
        return response



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class CoinbaseOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.coinbase.CoinbaseOAuth2"
    user_data_url = "https://api.coinbase.com/v2/user"
    expected_username = "satoshi_nakomoto"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "data": {
                "id": "1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa",
                "name": "Satoshi Nakamoto",
                "username": "satoshi_nakomoto",
                "profile_location": None,
                "profile_bio": None,
                "profile_url": "https://coinbase.com/satoshi_nakomoto",
                "avatar_url": None,
                "resource": "user",
                "resource_path": "/v2/user",
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Live OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/live.html
"""

from typing import Any

from .oauth import BaseOAuth2


class LiveOAuth2(BaseOAuth2):
    name = "live"
    AUTHORIZATION_URL = "https://login.live.com/oauth20_authorize.srf"
    ACCESS_TOKEN_URL = "https://login.live.com/oauth20_token.srf"
    SCOPE_SEPARATOR = ","
    DEFAULT_SCOPE = ["wl.basic", "wl.emails"]
    EXTRA_DATA = [
        ("id", "id"),
        ("access_token", "access_token"),
        ("authentication_token", "authentication_token"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
        ("email", "email"),
        ("first_name", "first_name"),
        ("last_name", "last_name"),
        ("token_type", "token_type"),
    ]
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Live Connect account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("first_name"), last_name=response.get("last_name")
        )
        return {
            "username": response.get("name"),
            "email": response.get("emails", {}).get("account", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://apis.live.net/v5.0/me", params={"access_token": access_token}
        )



"""
Bungie OAuth2 backend
"""

from typing import Any

from social_core.backends.oauth import BaseOAuth2


class BungieOAuth2(BaseOAuth2):
    name = "bungie"
    ID_KEY = "membership_id"
    AUTHORIZATION_URL = "https://www.bungie.net/en/oauth/authorize/"
    ACCESS_TOKEN_URL = "https://www.bungie.net/platform/app/oauth/token/"
    REFRESH_TOKEN_URL = "https://www.bungie.net/platform/app/oauth/token/"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("access_token", "access_token", True),
        ("expires_in", "expires_in"),
        ("membership_id", "membership_id"),
        ("refresh_expires_in", "refresh_expires_in"),
    ]

    def auth_headers(self):
        """Adds X-API-KEY and Origin"""
        return {
            "X-API-KEY": self.setting("API_KEY"),
            "Content-Type": "application/x-www-form-urlencoded",
            "Origin": self.setting("ORIGIN"),
            "Accept": "application/json",
        }

    def make_bungie_request(self, url, access_token, kwargs):
        """Helper function to get username data keyed off displayName"""
        headers = self.auth_headers()
        headers["Authorization"] = f"Bearer {access_token}"
        return self.get_json(url, headers=headers)

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()
        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], *args, response=response, **kwargs
        )

    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        data = self.user_data(access_token, *args, **kwargs)
        response = kwargs.get("response") or {}
        response.update(data or {})
        if "access_token" not in response:
            response["Response"]["access_token"]["value"] = access_token
        kwargs.update({"response": response, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Grab user profile information from Bunige"""
        membership_id = kwargs["response"]["membership_id"]
        url = "https://www.bungie.net/Platform/User/GetBungieNetUser/"
        response = self.make_bungie_request(url, access_token, kwargs)
        username = response["Response"]["user"]["displayName"]
        return {"username": username, "uid": membership_id}

    def get_user_details(self, response, *args, **kwargs):
        """Return user details from Bungie account"""
        username = response["username"]
        return {
            "first_name": username,
            "username": username,
            "uid": response["uid"],
        }



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class GrafanaOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.grafana.GrafanaOAuth2"
    user_data_url = "https://grafana.com/api/oauth2/user"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
        }
    )
    user_data_body = json.dumps(
        {"login": "fooboy", "email": "foo@bar.com", "name": "Foo Bar"}
    )
    expected_username = "fooboy"

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



from typing import Any
from urllib.parse import urlencode

from .oauth import BaseOAuth2


class WLCGOAuth2(BaseOAuth2):
    """
    WLCG IAM Authentication Backend
    """

    name = "wlcg"
    API_URL = "https://wlcg.cloud.cnaf.infn.it"
    AUTHORIZATION_URL = "https://wlcg.cloud.cnaf.infn.it/authorize"
    ACCESS_TOKEN_URL = "https://wlcg.cloud.cnaf.infn.it/token"
    REFRESH_TOKEN_URL = "https://wlcg.cloud.cnaf.infn.it/token"
    DEFAULT_SCOPE = ["openid", "email", "profile", "wlcg", "offline_access"]
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from WLCG IAM service"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("given_name"), last_name=response.get("family_name")
        )
        return {
            "username": response.get("email"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        url = f"https://wlcg.cloud.cnaf.infn.it/userinfo?{urlencode({'access_token': access_token})}"
        return self.get_json(url)



"""
RunKeeper OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/runkeeper.html
"""

from typing import Any

from .oauth import BaseOAuth2


class RunKeeperOAuth2(BaseOAuth2):
    """RunKeeper OAuth authentication backend"""

    name = "runkeeper"
    AUTHORIZATION_URL = "https://runkeeper.com/apps/authorize"
    ACCESS_TOKEN_URL = "https://runkeeper.com/apps/token"
    EXTRA_DATA = [
        ("userID", "id"),
    ]

    def get_user_id(self, details, response):
        return response["userID"]

    def get_user_details(self, response):
        """Parse username from profile link"""
        username = None
        profile_url = response.get("profile")
        if len(profile_url):
            profile_url_parts = profile_url.split("http://runkeeper.com/user/")
            if len(profile_url_parts) > 1 and len(profile_url_parts[1]):
                username = profile_url_parts[1]
        fullname, first_name, last_name = self.get_user_names(
            fullname=response.get("name")
        )
        return {
            "username": username,
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        # We need to use the /user endpoint to get the user id, the /profile
        # endpoint contains name, user name, location, gender
        user_data = self._user_data(access_token, "/user")
        profile_data = self._user_data(access_token, "/profile")
        return dict(user_data, **profile_data)

    def _user_data(self, access_token, path):
        url = f"https://api.runkeeper.com{path}"
        return self.get_json(url, params={"access_token": access_token})



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class UpworkOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.upwork.UpworkOAuth"
    user_data_url = "https://www.upwork.com/api/auth/v1/info.json"
    expected_username = "10101010"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "info": {
                "portrait_32_img": "",
                "capacity": {
                    "buyer": "no",
                    "affiliate_manager": "no",
                    "provider": "yes",
                },
                "company_url": "",
                "has_agency": "1",
                "portrait_50_img": "",
                "portrait_100_img": "",
                "location": {"city": "New York", "state": "", "country": "USA"},
                "ref": "9755314",
                "profile_url": "https://www.upwork.com/users/~10101010",
            },
            "auth_user": {
                "timezone": "USA/New York",
                "first_name": "Foo",
                "last_name": "Bar",
                "timezone_offset": "10000",
            },
            "server_time": "1111111111",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



from __future__ import annotations

from typing import Generic, TypeVar, cast
from unittest.mock import patch
from urllib.parse import urlparse

import requests
import responses

from social_core.backends.oauth import BaseOAuth1, BaseOAuth2, OAuthAuth
from social_core.tests.models import User
from social_core.utils import get_querystring, parse_qs, url_add_parameters

from .base import BaseBackendTest

OAuthBackendT = TypeVar("OAuthBackendT", bound=OAuthAuth)


class BaseOAuthTest(BaseBackendTest[OAuthBackendT], Generic[OAuthBackendT]):
    user_data_body: str | None = None
    user_data_url: str = ""
    user_data_url_post: bool = False
    user_data_content_type: str = "application/json"
    access_token_body: str | None = None
    access_token_status: int = 200

    def extra_settings(self) -> dict[str, str | list[str]]:
        assert self.name, "Subclasses must set the name attribute"
        return {
            f"SOCIAL_AUTH_{self.name}_KEY": "a-key",
            f"SOCIAL_AUTH_{self.name}_SECRET": "a-secret-key",
        }

    def _method(self, method):
        return {"GET": responses.GET, "POST": responses.POST}[method]

    def handle_state(self, start_url: str, target_url: str) -> str:
        start_query = parse_qs(urlparse(start_url).query)
        redirect_uri = start_query.get("redirect_uri")

        if getattr(self.backend, "STATE_PARAMETER", False) and start_query.get("state"):
            target_url = url_add_parameters(target_url, {"state": start_query["state"]})

        if redirect_uri and getattr(self.backend, "REDIRECT_STATE", False):
            redirect_query = parse_qs(urlparse(redirect_uri).query)
            if redirect_query.get("redirect_state"):
                target_url = url_add_parameters(
                    target_url, {"redirect_state": redirect_query["redirect_state"]}
                )
        return target_url

    def auth_handlers(self, start_url: str) -> str:
        target_url = self.handle_state(
            start_url, self.strategy.build_absolute_uri(self.complete_url)
        )
        responses.add(
            responses.GET,
            start_url,
            status=301,
            headers={"Location": target_url},
        )
        responses.add(responses.GET, target_url, status=200, body="foobar")
        if self.user_data_url:
            responses.add(
                responses.POST if self.user_data_url_post else responses.GET,
                self.user_data_url,
                body=self.user_data_body or "",
                content_type=self.user_data_content_type,
            )
        return target_url

    def pre_complete_callback(self, start_url) -> None:
        responses.add(
            self._method(self.backend.ACCESS_TOKEN_METHOD),
            url=self.backend.access_token_url(),
            status=self.access_token_status,
            body=self.access_token_body or "",
            content_type="application/json",
        )

    def do_start(self):
        start_url = self.backend.start().url
        target_url = self.auth_handlers(start_url)
        response = requests.get(start_url, timeout=1)
        self.assertEqual(response.url, target_url)
        self.assertEqual(response.text, "foobar")
        self.strategy.set_request_data(
            parse_qs(urlparse(start_url).query), self.backend
        )
        self.strategy.set_request_data(
            parse_qs(urlparse(target_url).query), self.backend
        )
        self.pre_complete_callback(start_url)
        return self.backend.complete()


BaseOAuth1BackendT = TypeVar("BaseOAuth1BackendT", bound=BaseOAuth1)


class OAuth1Test(BaseOAuthTest[BaseOAuth1BackendT], Generic[BaseOAuth1BackendT]):
    request_token_body: str
    raw_complete_url = "/complete/{0}/?oauth_verifier=bazqux&oauth_token=foobar"

    def request_token_handler(self) -> None:
        assert self.request_token_body, "Subclasses must set request_token_body"
        responses.add(
            self._method(self.backend.REQUEST_TOKEN_METHOD),
            self.backend.REQUEST_TOKEN_URL,
            body=self.request_token_body,
            status=200,
        )

    def do_start(self):
        self.request_token_handler()
        return super().do_start()


BaseOAuth2BackendT = TypeVar("BaseOAuth2BackendT", bound=BaseOAuth2)


class OAuth2Test(BaseOAuthTest[BaseOAuth2BackendT], Generic[BaseOAuth2BackendT]):
    raw_complete_url = "/complete/{0}/?code=foobar"
    refresh_token_body = ""

    def refresh_token_arguments(self):
        return {}

    def do_refresh_token(self):
        self.do_login()
        responses.add(
            self._method(self.backend.REFRESH_TOKEN_METHOD),
            self.backend.refresh_token_url(),
            status=200,
            body=self.refresh_token_body,
        )
        user = next(iter(User.cache.values()))
        social = user.social[0]
        social.refresh_token(strategy=self.strategy, **self.refresh_token_arguments())
        return user, social


class OAuth2PkcePlainTest(OAuth2Test):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {f"SOCIAL_AUTH_{self.name}_PKCE_CODE_CHALLENGE_METHOD": "plain"}
        )
        return settings

    def do_login(self):
        user = super().do_login()

        auth_request = next(
            r.request
            for r in responses.calls
            if cast("str", r.request.url).startswith(self.backend.authorization_url())
        )
        code_challenge = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge"
        )
        code_challenge_method = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge_method"
        )
        self.assertIsNotNone(code_challenge)
        self.assertEqual(code_challenge_method, "plain")

        auth_complete = next(
            r.request
            for r in responses.calls
            if cast("str", r.request.url).startswith(self.backend.access_token_url())
        )
        code_verifier = parse_qs(auth_complete.body).get("code_verifier")
        self.assertEqual(code_challenge, code_verifier)

        return user


class OAuth2PkceS256Test(OAuth2Test):
    def do_login(self):
        # use default value of PKCE_CODE_CHALLENGE_METHOD (s256)
        user = super().do_login()

        auth_request = next(
            r.request
            for r in responses.calls
            if cast("str", r.request.url).startswith(self.backend.authorization_url())
        )
        code_challenge = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge"
        )
        code_challenge_method = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge_method"
        )
        self.assertIsNotNone(code_challenge)
        self.assertIn(code_challenge_method, ["s256", "S256"])

        auth_complete = next(
            r.request
            for r in responses.calls
            if cast("str", r.request.url).startswith(self.backend.access_token_url())
        )
        code_verifier = parse_qs(auth_complete.body).get("code_verifier")
        self.assertEqual(
            self.backend.generate_code_challenge(code_verifier, code_challenge_method),
            code_challenge,
        )

        return user


class BaseAuthUrlTestMixin(Generic[OAuthBackendT]):
    backend: OAuthBackendT

    def check_parameters_in_authorization_url(
        self, auth_url_key="AUTHORIZATION_URL"
    ) -> None:
        """
        Check the parameters in authorization url

        When inserting parameters directly into AUTHORIZATION_URL, we expect the
        other parameters to be added to the end of the url
        """
        original_url = (
            self.backend.AUTHORIZATION_URL or self.backend.authorization_url()
        )
        with (
            patch.object(
                self.backend,
                "authorization_url",
                return_value=f"{original_url}?param1=value1&param2=value2",
            ),
            patch.object(
                self.backend,
                auth_url_key,
                f"{original_url}?param1=value1&param2=value2",
            ),
        ):
            # we expect an & symbol to join the different parameters
            assert "?param1=value1&param2=value2&" in self.backend.auth_url()

    def test_auth_url_parameters(self) -> None:
        self.check_parameters_in_authorization_url()


class OAuth1AuthUrlTestMixin(BaseAuthUrlTestMixin):
    def test_auth_url_parameters(self) -> None:
        self.request_token_handler()  # type: ignore[attr-defined]
        self.check_parameters_in_authorization_url()



from __future__ import annotations

from typing import TYPE_CHECKING

from social_core.exceptions import NotAllowedToDisconnect

if TYPE_CHECKING:
    from social_core.storage import UserProtocol
    from social_core.strategy import BaseStrategy


def allowed_to_disconnect(
    strategy: BaseStrategy,
    user: UserProtocol | None,
    name,
    user_storage,
    association_id=None,
    *args,
    **kwargs,
) -> None:
    if not user_storage.allowed_to_disconnect(user, name, association_id):
        raise NotAllowedToDisconnect


def get_entries(
    strategy: BaseStrategy,
    user: UserProtocol | None,
    name,
    user_storage,
    association_id=None,
    *args,
    **kwargs,
):
    return {
        "entries": user_storage.get_social_auth_for_user(user, name, association_id)
    }


def revoke_tokens(strategy, entries, *args, **kwargs) -> None:
    revoke_tokens_on_disconnect = strategy.setting("REVOKE_TOKENS_ON_DISCONNECT", False)
    if revoke_tokens_on_disconnect:
        for entry in entries:
            if "access_token" in entry.extra_data:
                backend = entry.get_backend_instance(strategy)
                backend.revoke_token(entry.extra_data["access_token"], entry.uid)


def disconnect(strategy: BaseStrategy, entries, user_storage, *args, **kwargs) -> None:
    for entry in entries:
        user_storage.disconnect(entry)



"""
Clef OAuth support.

This contribution adds support for Clef OAuth service. The settings
SOCIAL_AUTH_CLEF_KEY and SOCIAL_AUTH_CLEF_SECRET must be defined with the
values given by Clef application registration process.
"""

from typing import Any

from .oauth import BaseOAuth2


class ClefOAuth2(BaseOAuth2):
    """Clef OAuth authentication backend"""

    name = "clef"
    AUTHORIZATION_URL = "https://clef.io/iframes/qr"
    ACCESS_TOKEN_URL = "https://clef.io/api/v1/authorize"
    SCOPE_SEPARATOR = ","

    def auth_params(self, *args, **kwargs):
        params = super().auth_params(*args, **kwargs)
        params["app_id"] = params.pop("client_id")
        params["redirect_url"] = params.pop("redirect_uri")
        return params

    def get_user_details(self, response):
        """Return user details from Github account"""
        info = response.get("info")
        fullname, first_name, last_name = self.get_user_names(
            first_name=info.get("first_name"), last_name=info.get("last_name")
        )

        email = info.get("email", "")
        username = email.split("@", 1)[0] if email else info.get("id")

        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "phone_number": info.get("phone_number", ""),
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.get_json(
            "https://clef.io/api/v1/info", params={"access_token": access_token}
        )



from typing import Any

from social_core.backends.oauth import BaseOAuth2


class AtlassianOAuth2(BaseOAuth2):
    name = "atlassian"
    AUTHORIZATION_URL = "https://auth.atlassian.com/authorize"
    ACCESS_TOKEN_URL = "https://auth.atlassian.com/oauth/token"
    DEFAULT_SCOPE = ["read:jira-user", "offline_access"]
    ID_KEY = "accountId"
    EXTRA_DATA = [
        ("resources", "resources"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
    ]

    def auth_params(self, state=None):
        params = super().auth_params(state)
        params.update({"audience": "api.atlassian.com", "prompt": "consent"})
        return params

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(response["displayName"])
        return {
            "username": response["accountId"],
            "email": response["emailAddress"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        resources = self.get_json(
            "https://api.atlassian.com/oauth/token/accessible-resources",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        user_info = self.get_json(
            f"https://api.atlassian.com/ex/jira/{resources[0]['id']}/rest/api/2/myself",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        user_info["resources"] = resources
        return user_info
